#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Name class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .secondaryobj import SecondaryObject
from .privacybase import PrivacyBase
from .citationbase import CitationBase
from .notebase import NoteBase
from .datebase import DateBase
from .surnamebase import SurnameBase
from .nametype import NameType
from .const import IDENTICAL, EQUAL, DIFFERENT
from .date import Date
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Personal Name
#
# -------------------------------------------------------------------------
class Name(SecondaryObject, PrivacyBase, SurnameBase, CitationBase, NoteBase, DateBase):
    """
    Provide name information about a person.

    A person may have more that one name throughout his or her life. The Name
    object stores one of them
    """

    DEF = 0  # Default format (determined by gramps-wide prefs)
    LNFN = 1  # last name first name
    FNLN = 2  # first name last name
    FN = 4  # first name
    LNFNP = 5  # primary name primconnector rest, given pa/ma suffix, primprefix

    NAMEFORMATS = (DEF, LNFN, FNLN, FN, LNFNP)
    # deprecated :
    PTFN = 3  # patronymic first name

    def __init__(self, source=None, data=None):
        """Create a new Name instance, copying from the source if provided.
        We should connect here to 'person-groupname-rebuild' and do something
        correct when first parameter is the name, and second parameter is
        different from the group here. However, that would be complicated and
        no real errors that cannot be ammended can be done if group is
        saved differently.
        """
        PrivacyBase.__init__(self, source)
        SurnameBase.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        DateBase.__init__(self, source)
        if data:
            (
                privacy,
                citation_list,
                note,
                date,
                self.first_name,
                surname_list,
                self.suffix,
                self.title,
                name_type,
                self.group_as,
                self.sort_as,
                self.display_as,
                self.call,
                self.nick,
                self.famnick,
            ) = data
            self.type = NameType(name_type)
            SurnameBase.unserialize(self, surname_list)
            PrivacyBase.unserialize(self, privacy)
            CitationBase.unserialize(self, citation_list)
            NoteBase.unserialize(self, note)
            DateBase.unserialize(self, date)
        elif source:
            self.first_name = source.first_name
            self.suffix = source.suffix
            self.title = source.title
            self.type = NameType(source.type)
            self.group_as = source.group_as
            self.sort_as = source.sort_as
            self.display_as = source.display_as
            self.call = source.call
            self.nick = source.nick
            self.famnick = source.famnick
        else:
            self.first_name = ""
            self.suffix = ""
            self.title = ""
            self.type = NameType()
            self.group_as = ""
            self.sort_as = self.DEF
            self.display_as = self.DEF
            self.call = ""
            self.nick = ""
            self.famnick = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            PrivacyBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            DateBase.serialize(self),
            self.first_name,
            SurnameBase.serialize(self),
            self.suffix,
            self.title,
            self.type.serialize(),
            self.group_as,
            self.sort_as,
            self.display_as,
            self.call,
            self.nick,
            self.famnick,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        from .surname import Surname

        return {
            "type": "object",
            "title": _("Name"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "private": {"type": "boolean", "title": _("Private")},
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "date": {
                    "oneOf": [{"type": "null"}, Date.get_schema()],
                    "title": _("Date"),
                },
                "first_name": {"type": "string", "title": _("Given name")},
                "surname_list": {
                    "type": "array",
                    "items": Surname.get_schema(),
                    "title": _("Surnames"),
                },
                "suffix": {"type": "string", "title": _("Suffix")},
                "title": {"type": "string", "title": _("Title")},
                "type": NameType.get_schema(),
                "group_as": {"type": "string", "title": _("Group as")},
                "sort_as": {"type": "integer", "title": _("Sort as")},
                "display_as": {"type": "integer", "title": _("Display as")},
                "call": {"type": "string", "title": _("Call name")},
                "nick": {"type": "string", "title": _("Nick name")},
                "famnick": {"type": "string", "title": _("Family nick name")},
            },
        }

    def is_empty(self):
        """
        Indicate if the name is empty.
        """
        namefieldsempty = (
            self.first_name == ""
            and self.suffix == ""
            and self.title == ""
            and self.nick == ""
            and self.famnick == ""
        )
        surnamefieldsempty = False not in [
            surn.is_empty() for surn in self.surname_list
        ]
        return namefieldsempty and surnamefieldsempty

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (
            privacy,
            citation_list,
            note_list,
            date,
            self.first_name,
            surname_list,
            self.suffix,
            self.title,
            name_type,
            self.group_as,
            self.sort_as,
            self.display_as,
            self.call,
            self.nick,
            self.famnick,
        ) = data
        self.type = NameType(name_type)
        PrivacyBase.unserialize(self, privacy)
        SurnameBase.unserialize(self, surname_list)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        DateBase.unserialize(self, date)
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [
            self.first_name,
            self.suffix,
            self.title,
            str(self.type),
            self.call,
            self.nick,
            self.famnick,
        ]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.surname_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return []

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return (
            self.get_referenced_note_handles() + self.get_referenced_citation_handles()
        )

    def is_equivalent(self, other):
        """
        Return if this name is equivalent, that is agrees in type, first,
        call, surname_list, suffix, title and date, to other.

        :param other: The name to compare this name to.
        :type other: Name
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        # TODO what to do with sort and display?
        if (
            self.get_text_data_list() != other.get_text_data_list()
            or self.get_date_object() != other.get_date_object()
            or SurnameBase.serialize(self) != SurnameBase.serialize(other)
        ):
            return DIFFERENT
        else:
            if self.is_equal(other):
                return IDENTICAL
            else:
                return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this name.
        Normally the person merge code should opt for adding an alternate
        name if names are actually different (like not equal surname list)

        Lost: type, first, call, suffix, title, nick, famnick and date of
        acquisition.

        :param acquisition: The name to merge with the present name.
        :type acquisition: Name
        """
        # TODO what to do with sort and display?
        self._merge_privacy(acquisition)
        self._merge_surname_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)

    def set_group_as(self, name):
        """
        Set the grouping name for a person.

        Normally, this is the person's surname. However, some locales group
        equivalent names (e.g. Ivanova and Ivanov in Russian are usually
        considered equivalent.

        .. note:: There is also a database wide grouping set_name_group_mapping
          So one might map a name Smith to SmithNew, and have one person still
          grouped with name Smith. Hence, group_as can be equal to surname!
        """
        self.group_as = name

    def get_group_as(self):
        """
        Return the grouping name, which is used to group equivalent surnames.
        """
        return self.group_as

    def get_group_name(self):
        """
        Return the grouping name, which is used to group equivalent surnames.
        """
        if self.group_as:
            return self.group_as
        else:
            return self.get_primary_surname().get_surname()

    def set_sort_as(self, value):
        """
        Specifies the sorting method for the specified name.

        Typically the locale's default should be used. However, there may be
        names where a specific sorting structure is desired for a name.
        """
        self.sort_as = value

    def get_sort_as(self):
        """
        Return the selected sorting method for the name.

        The options are LNFN (last name, first name), FNLN (first name, last
        name), etc.
        """
        return self.sort_as

    def set_display_as(self, value):
        """
        Specifies the display format for the specified name.

        Typically the locale's default should be used. However, there may be
        names where a specific display format is desired for a name.
        """
        self.display_as = value

    def get_display_as(self):
        """
        Return the selected display format for the name.

        The options are LNFN (last name, first name), FNLN (first name, last
        name), etc.
        """
        return self.display_as

    def get_call_name(self):
        """
        Return the call name.

        The call name's exact definition is not predetermined, and may be
        locale specific.
        """
        return self.call

    def set_call_name(self, val):
        """
        Set the call name.

        The call name's exact definition is not predetermined, and may be
        locale specific.
        """
        self.call = val

    def get_nick_name(self):
        """
        Return the nick name.

        The nick name of the person, a not official name the person is known
        with.
        """
        return self.nick

    def set_nick_name(self, val):
        """
        Set the nick name.

        The nick name of the person, a not official name the person is known
        with.
        """
        self.nick = val

    def get_family_nick_name(self):
        """
        Return the family nick name.

        The family nick name of the family of the person, a not official name
        use to denote the entire family.
        """
        return self.famnick

    def set_family_nick_name(self, val):
        """
        Set the family nick name.

        The family nick name of the family of the person, a not official name
        use to denote the entire family.
        """
        self.famnick = val

    def set_type(self, the_type):
        """Set the type of the Name instance."""
        self.type.set(the_type)

    def get_type(self):
        """Return the type of the Name instance."""
        return self.type

    def set_first_name(self, name):
        """Set the given name for the Name instance."""
        self.first_name = name

    def get_first_name(self):
        """Return the given name for the Name instance."""
        return self.first_name

    def set_suffix(self, name):
        """Set the suffix (such as Jr., III, etc.) for the Name instance."""
        self.suffix = name

    def get_suffix(self):
        """Return the suffix for the Name instance."""
        return self.suffix

    def set_title(self, title):
        """Set the title (Dr., Reverand, Captain) for the Name instance."""
        self.title = title

    def get_title(self):
        """Return the title for the Name instance."""
        return self.title

    def get_name(self):
        """
        Return a name string built from the components of the Name instance,
        in the form of: surname, Firstname.
        """
        first = self.first_name
        surname = self.get_surname()
        if self.suffix:
            # Translators: needed for Arabic, ignore otherwise
            return _("%(surname)s, %(first)s %(suffix)s") % {
                "surname": surname,
                "first": first,
                "suffix": self.suffix,
            }
        else:
            # Translators: needed for Arabic, ignore otherwise
            return _("%(str1)s, %(str2)s") % {"str1": surname, "str2": first}

    def get_upper_name(self):
        """
        Return a name string built from the components of the Name instance,
        in the form of SURNAME, Firstname.
        """
        first = self.first_name
        surname = self.get_surname().upper()
        if self.suffix:
            # Translators: needed for Arabic, ignore otherwise
            return _("%(surname)s, %(first)s %(suffix)s") % {
                "surname": surname,
                "first": first,
                "suffix": self.suffix,
            }
        else:
            # Translators: needed for Arabic, ignore otherwise
            return _("%(str1)s, %(str2)s") % {"str1": surname, "str2": first}

    def get_regular_name(self):
        """
        Return a name string built from the components of the Name instance,
        in the form of Firstname surname.
        """
        first = self.first_name
        surname = self.get_surname()
        if self.suffix == "":
            return "%s %s" % (first, surname)
        else:
            # Translators: needed for Arabic, ignore otherwise
            return _("%(first)s %(surname)s, %(suffix)s") % {
                "surname": surname,
                "first": first,
                "suffix": self.suffix,
            }

    def get_gedcom_parts(self):
        """
        Returns a GEDCOM-formatted name dictionary.

        .. note:: Fields patronymic and prefix are deprecated, prefix_list and
                  surname list, added.
        """
        retval = {}
        retval["given"] = self.first_name.strip()
        retval["surname"] = self.get_surname().replace("/", "?")
        retval["suffix"] = self.suffix
        retval["title"] = self.title
        retval["surnamelist"] = self.get_surnames()
        retval["prefixes"] = self.get_prefixes()
        retval["connectors"] = self.get_connectors()
        retval["nick"] = self.nick
        retval["famnick"] = self.famnick
        return retval

    def get_gedcom_name(self):
        """
        Returns a GEDCOM-formatted name.
        """
        firstname = self.first_name.strip()
        surname = self.get_surname().replace("/", "?")
        suffix = self.suffix
        if suffix == "":
            return "%s /%s/" % (firstname, surname)
        else:
            return "%s /%s/ %s" % (firstname, surname, suffix)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2012       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman

"Rebuild gender stat values"

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

log = logging.getLogger(".RebuildGenderStat")

# -------------------------------------------------------------------------
#
# gtk modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.plug import tool
from gramps.gui.dialog import OkDialog
from gramps.gen.updatecallback import UpdateCallback
from gramps.gen.lib import Name

# -------------------------------------------------------------------------
#
# runTool
#
# -------------------------------------------------------------------------

COLUMN_GENDER = 2
COLUMN_NAME = 3
COLUMN_ALTNAMES = 4


class RebuildGenderStat(tool.Tool, UpdateCallback):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate

        tool.Tool.__init__(self, dbstate, options_class, name)

        if self.db.readonly:
            return

        self.db.disable_signals()
        if uistate:
            self.callback = uistate.pulse_progressbar
            uistate.set_busy_cursor(True)
            uistate.progress.show()
            uistate.push_message(
                dbstate, _("Rebuilding gender statistics for name gender guessing...")
            )
        else:
            self.callback = None
            print("Rebuilding gender statistics for name gender guessing...")

        UpdateCallback.__init__(self, self.callback)
        self.set_total(self.db.get_number_of_people())
        self.rebuild_genderstats()
        self.reset()

        if uistate:
            uistate.set_busy_cursor(False)
            uistate.progress.hide()
            OkDialog(
                _("Gender statistics rebuilt"),
                _("Gender statistics for name gender guessing have been rebuilt."),
                parent=uistate.window,
            )
        else:
            print("Gender statistics for name gender guessing have been rebuilt.")
        self.db.enable_signals()

    def rebuild_genderstats(self):
        """
        Function to rebuild the gender stats
        """
        self.db.genderStats.clear_stats()
        with self.db.get_person_cursor() as cursor:
            # loop over database and store the sort field, and the handle, and
            # allow for a third iter
            for key, data in cursor:
                rawprimname = data[COLUMN_NAME]
                rawaltnames = data[COLUMN_ALTNAMES]
                primary_name = Name().unserialize(rawprimname).get_first_name()
                alternate_names = [
                    Name().unserialize(name).get_first_name() for name in rawaltnames
                ]
                self.db.genderStats.count_name(primary_name, data[COLUMN_GENDER])


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class RebuildGenderStatOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  Douglas S. Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import time

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------

from gramps.gen.lib import Person, Family
from gramps.gen.db import PERSON_KEY, FAMILY_KEY, TXNDEL
from gramps.gen.plug import Gramplet
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.utils.db import family_name
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# ------------------------------------------------------------------------
#
# LogGramplet class
#
# ------------------------------------------------------------------------
class LogGramplet(Gramplet):
    def init(self):
        self.set_tooltip(_("Click name to change active\nDouble-click name to edit"))
        self.set_text(_("Log for this Session") + "\n")
        self.gui.force_update = True  # will always update, even if minimized
        self.last_log = None

    def timestamp(self):
        self.append_text(time.strftime("%Y-%m-%d %H:%M:%S "))

    def db_changed(self):
        self.timestamp()
        self.append_text(_("Opened data base -----------\n"))
        # List of translated strings used here (translated in self.log ).
        _("Added"), _("Deleted"), _("Edited"), _("Selected")  # Dead code for l10n
        self.connect(
            self.dbstate.db,
            "person-add",
            lambda handles: self.log("Person", "Added", handles),
        )
        self.connect(
            self.dbstate.db,
            "person-delete",
            lambda handles: self.log("Person", "Deleted", handles),
        )
        self.connect(
            self.dbstate.db,
            "person-update",
            lambda handles: self.log("Person", "Edited", handles),
        )
        self.connect(
            self.dbstate.db,
            "family-add",
            lambda handles: self.log("Family", "Added", handles),
        )
        self.connect(
            self.dbstate.db,
            "family-delete",
            lambda handles: self.log("Family", "Deleted", handles),
        )
        self.connect(
            self.dbstate.db,
            "family-update",
            lambda handles: self.log("Family", "Edited", handles),
        )
        self.connect_signal("Person", self.active_changed)
        self.connect_signal("Family", self.active_changed_family)

    def active_changed(self, handle):
        if handle:
            self.log("Person", "Selected", [handle])

    def active_changed_family(self, handle):
        if handle:
            self.log("Family", "Selected", [handle])

    def log(self, ltype, action, handles):
        for handle in set(handles):
            if self.last_log == (ltype, action, handle):
                continue
            self.last_log = (ltype, action, handle)
            self.timestamp()
            # Translators: needed for French, ignore otherwise
            self.append_text(_("%s: ") % _(action))
            if action == "Deleted":
                transaction = self.dbstate.db.transaction
                if ltype == "Person":
                    name = "a person"
                    if transaction is not None:
                        for i in transaction.get_recnos(reverse=True):
                            (
                                obj_type,
                                trans_type,
                                hndl,
                                old_data,
                                dummy,
                            ) = transaction.get_record(i)
                            if isinstance(hndl, bytes):
                                hndl = str(hndl, "utf-8")
                            if (
                                obj_type == PERSON_KEY
                                and trans_type == TXNDEL
                                and hndl == handle
                            ):
                                person = Person()
                                person.unserialize(old_data)
                                name = name_displayer.display(person)
                                break
                elif ltype == "Family":
                    name = "a family"
                    if transaction is not None:
                        for i in transaction.get_recnos(reverse=True):
                            (
                                obj_type,
                                trans_type,
                                hndl,
                                old_data,
                                dummy,
                            ) = transaction.get_record(i)
                            if isinstance(hndl, bytes):
                                hndl = str(hndl, "utf-8")
                            if (
                                obj_type == FAMILY_KEY
                                and trans_type == TXNDEL
                                and hndl == handle
                            ):
                                family = Family()
                                family.unserialize(old_data)
                                name = family_name(family, self.dbstate.db, name)
                                break
                self.append_text(name)
            else:
                if ltype == "Person":
                    person = self.dbstate.db.get_person_from_handle(handle)
                    name = name_displayer.display(person)
                elif ltype == "Family":
                    family = self.dbstate.db.get_family_from_handle(handle)
                    name = family_name(family, self.dbstate.db, "a family")
                self.link(name, ltype, handle)
            self.append_text("\n")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# gtk
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ... import widgets
from gramps.gen.lib import Repository, RepositoryType
from .. import build_filter_model
from . import SidebarFilter
from gramps.gen.filters import GenericFilterFactory, rules
from gramps.gen.filters.rules.repository import (
    RegExpIdOf,
    HasRepo,
    HasTag,
    HasNoteRegexp,
    MatchesFilter,
)

GenericRepoFilter = GenericFilterFactory("Repository")


# -------------------------------------------------------------------------
#
# RepoSidebarFilter class
#
# -------------------------------------------------------------------------
class RepoSidebarFilter(SidebarFilter):
    def __init__(self, dbstate, uistate, clicked):
        self.clicked_func = clicked
        self.filter_id = widgets.BasicEntry()
        self.filter_title = widgets.BasicEntry()
        self.filter_address = widgets.BasicEntry()
        self.filter_url = widgets.BasicEntry()
        self.repo = Repository()
        self.repo.set_type((RepositoryType.CUSTOM, ""))
        self.rtype = Gtk.ComboBox(has_entry=True)
        if dbstate.is_open():
            self.custom_types = dbstate.db.get_event_types()
        else:
            self.custom_types = []

        self.event_menu = widgets.MonitoredDataType(
            self.rtype,
            self.repo.set_type,
            self.repo.get_type,
            False,  # read-only?
            self.custom_types,
        )

        self.filter_note = widgets.BasicEntry()

        self.filter_regex = Gtk.CheckButton(label=_("Use regular expressions"))
        self.sensitive_regex = Gtk.CheckButton(label=_("Case sensitive"))

        self.tag = Gtk.ComboBox()
        self.generic = Gtk.ComboBox()

        SidebarFilter.__init__(self, dbstate, uistate, "Repository")

    def create_widget(self):
        cell = Gtk.CellRendererText()
        cell.set_property("width", self._FILTER_WIDTH)
        cell.set_property("ellipsize", self._FILTER_ELLIPSIZE)
        self.generic.pack_start(cell, True)
        self.generic.add_attribute(cell, "text", 0)
        self.on_filters_changed("Repository")

        cell = Gtk.CellRendererText()
        cell.set_property("width", self._FILTER_WIDTH)
        cell.set_property("ellipsize", self._FILTER_ELLIPSIZE)
        self.tag.pack_start(cell, True)
        self.tag.add_attribute(cell, "text", 0)

        self.rtype.get_child().set_width_chars(5)

        self.add_text_entry(_("ID"), self.filter_id)
        self.add_text_entry(_("Name"), self.filter_title)
        self.add_entry(_("Type"), self.rtype)
        self.add_text_entry(_("Address"), self.filter_address)
        self.add_text_entry(_("URL"), self.filter_url)
        self.add_text_entry(_("Note"), self.filter_note)
        self.add_entry(_("Tag"), self.tag)
        self.add_filter_entry(_("Custom filter"), self.generic)
        self.add_regex_entry(self.filter_regex)
        self.add_regex_case(self.sensitive_regex)

    def clear(self, obj):
        self.filter_id.set_text("")
        self.filter_title.set_text("")
        self.filter_address.set_text("")
        self.filter_url.set_text("")
        self.rtype.get_child().set_text("")
        self.filter_note.set_text("")
        self.tag.set_active(0)
        self.generic.set_active(0)

    def get_filter(self):
        gid = str(self.filter_id.get_text()).strip()
        title = str(self.filter_title.get_text()).strip()
        address = str(self.filter_address.get_text()).strip()
        url = str(self.filter_url.get_text()).strip()
        rtype = self.repo.get_type().xml_str()
        note = str(self.filter_note.get_text()).strip()
        regex = self.filter_regex.get_active()
        usecase = self.sensitive_regex.get_active()
        tag = self.tag.get_active() > 0
        gen = self.generic.get_active() > 0

        empty = not (
            gid or title or address or url or rtype or note or regex or tag or gen
        )
        if empty:
            generic_filter = None
        else:
            generic_filter = GenericRepoFilter()
            if gid:
                rule = RegExpIdOf([gid], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            rule = HasRepo(
                [title, rtype, address, url], use_regex=regex, use_case=usecase
            )
            generic_filter.add_rule(rule)

            if note:
                rule = HasNoteRegexp([note], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            # check the Tag
            if tag:
                model = self.tag.get_model()
                node = self.tag.get_active_iter()
                attr = model.get_value(node, 0)
                rule = HasTag([attr])
                generic_filter.add_rule(rule)

            if self.generic.get_active() != 0:
                model = self.generic.get_model()
                node = self.generic.get_active_iter()
                obj = str(model.get_value(node, 0))
                rule = MatchesFilter([obj])
                generic_filter.add_rule(rule)

        return generic_filter

    def on_filters_changed(self, name_space):
        if name_space == "Repository":
            all_filter = GenericRepoFilter()
            all_filter.set_name(_("None"))
            all_filter.add_rule(rules.repository.AllRepos([]))
            self.generic.set_model(build_filter_model("Repository", [all_filter]))
            self.generic.set_active(0)

    def on_tags_changed(self, tag_list):
        """
        Update the list of tags in the tag filter.
        """
        model = Gtk.ListStore(str)
        model.append(("",))
        for tag_name in tag_list:
            model.append((tag_name,))
        self.tag.set_model(model)
        self.tag.set_active(0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
#               2011  Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from copy import deepcopy

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.lib import EventType, NoteType
from gramps.gen.db import DbTxn
from ..glade import Glade
from .displaytabs import (
    CitationEmbedList,
    NoteTab,
    GalleryTab,
    EventBackRefList,
    EventAttrEmbedList,
)
from ..widgets import (
    PrivacyButton,
    MonitoredEntry,
    MonitoredDate,
    MonitoredDataType,
    MonitoredTagList,
)
from .editreference import RefTab, EditReference

from .objectentries import PlaceEntry

from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("Event_Reference_Editor_dialog", "manual")


# -------------------------------------------------------------------------
#
# EditEventRef class
#
# -------------------------------------------------------------------------
class EditEventRef(EditReference):
    def __init__(self, state, uistate, track, event, event_ref, update):
        EditReference.__init__(self, state, uistate, track, event, event_ref, update)
        self.original = deepcopy(event.serialize())
        self._init_event()

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel,
            self.top.get_object("eer_title"),
            _("Event Reference Editor"),
        )
        self.setup_configs("interface.event-ref", 600, 450)

        self.define_warn_box(self.top.get_object("eer_warning"))
        self.define_expander(self.top.get_object("eer_expander"))
        self.share_btn = self.top.get_object("share_place")
        self.add_del_btn = self.top.get_object("add_del_place")

        tblref = self.top.get_object("table64")
        notebook = self.top.get_object("notebook_ref")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.reftab = RefTab(
            self.dbstate, self.uistate, self.track, _("General"), tblref
        )
        tblref = self.top.get_object("table62")
        notebook = self.top.get_object("notebook")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.primtab = RefTab(
            self.dbstate, self.uistate, self.track, _("_General"), tblref
        )

    def _post_init(self):
        date = self.top.get_object("eer_date_entry")
        if not date.get_text_length():
            date.grab_focus()

    def _init_event(self):
        if not self.db.readonly:
            self.commit_event = self.db.commit_event
            self.add_event = self.db.add_event

    def get_custom_events(self):
        return sorted(self.db.get_event_types(), key=lambda s: s.lower())

    def _connect_signals(self):
        self.define_ok_button(self.top.get_object("ok"), self.ok_clicked)
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("event-rebuild", self.close)
        self._add_db_signal("event-delete", self.check_for_close)
        self._add_db_signal("place-delete", self.place_delete)
        self._add_db_signal("place-update", self.place_update)

    def _setup_fields(self):
        self.ref_privacy = PrivacyButton(
            self.top.get_object("eer_ref_priv"), self.source_ref, self.db.readonly
        )

        self.descr_field = MonitoredEntry(
            self.top.get_object("eer_description"),
            self.source.set_description,
            self.source.get_description,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.source.set_gramps_id,
            self.source.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.source.set_tag_list,
            self.source.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.place_field = PlaceEntry(
            self.dbstate,
            self.uistate,
            self.track,
            self.top.get_object("eer_place"),
            self.top.get_object("eer_place_event_box"),
            self.source.set_place_handle,
            self.source.get_place_handle,
            self.share_btn,
            self.add_del_btn,
        )

        self.ev_privacy = PrivacyButton(
            self.top.get_object("eer_ev_priv"), self.source, self.db.readonly
        )

        self.role_selector = MonitoredDataType(
            self.top.get_object("eer_role_combo"),
            self.source_ref.set_role,
            self.source_ref.get_role,
            self.db.readonly,
            self.db.get_event_roles(),
        )

        self.event_menu = MonitoredDataType(
            self.top.get_object("eer_type_combo"),
            self.source.set_type,
            self.source.get_type,
            self.db.readonly,
            custom_values=self.get_custom_events(),
        )

        self.date_check = MonitoredDate(
            self.top.get_object("eer_date_entry"),
            self.top.get_object("eer_date_stat"),
            self.source.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """

        notebook = self.top.get_object("notebook")
        notebook_ref = self.top.get_object("notebook_ref")

        self._add_tab(notebook, self.primtab)
        self._add_tab(notebook_ref, self.reftab)
        self.track_ref_for_deletion("primtab")
        self.track_ref_for_deletion("reftab")

        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_citation_list(),
            "eventref_editor_shared_citations",
        )
        self._add_tab(notebook, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.srcref_ref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_citation_list(),
            "eventref_editor_ref_citations",
        )
        self._add_tab(notebook_ref, self.srcref_ref_list)
        self.track_ref_for_deletion("srcref_ref_list")

        self.attr_list = EventAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_attribute_list(),
            "eventref_editor_shared_attributes",
        )
        self._add_tab(notebook, self.attr_list)
        self.track_ref_for_deletion("attr_list")

        self.attr_ref_list = EventAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_attribute_list(),
            "eventref_editor_ref_attributes",
        )
        self._add_tab(notebook_ref, self.attr_ref_list)
        self.track_ref_for_deletion("attr_ref_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_note_list(),
            "eventref_editor_shared_notes",
            notetype=NoteType.EVENT,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.note_ref_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_note_list(),
            "eventref_editor_ref_notes",
            notetype=NoteType.EVENTREF,
        )
        self._add_tab(notebook_ref, self.note_ref_tab)
        self.track_ref_for_deletion("note_ref_tab")

        self.gallery_tab = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_media_list(),
        )
        self._add_tab(notebook, self.gallery_tab)
        self.track_ref_for_deletion("gallery_tab")

        self.backref_tab = EventBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.source.handle),
            "eventref_editor_references",
            self.enable_warnbox,
        )
        self._add_tab(notebook, self.backref_tab)
        self.track_ref_for_deletion("backref_tab")

        self._setup_notebook_tabs(notebook)
        self._setup_notebook_tabs(notebook_ref)

    def build_menu_names(self, eventref):
        if self.source:
            event_name = str(self.source.get_type())
            submenu_label = _("Event: %s") % event_name
        else:
            submenu_label = _("New Event")
        return (_("Event Reference Editor"), submenu_label)

    def ok_clicked(self, obj):
        if self.source.handle:
            # only commit if it has changed
            if self.source.serialize() != self.original:
                with DbTxn(_("Modify Event"), self.db) as trans:
                    self.commit_event(self.source, trans)
        else:
            if self.check_for_duplicate_id("Event"):
                return
            with DbTxn(_("Add Event"), self.db) as trans:
                self.add_event(self.source, trans)
            self.source_ref.ref = self.source.handle

        if self.update:
            self.update(self.source_ref, self.source)

        self.close()

    def place_update(self, hndls):
        """Place changed outside of dialog, update text if its ours"""
        handle = self.source.get_place_handle()
        if handle and handle in hndls:
            place = self.db.get_place_from_handle(handle)
            p_lbl = "%s [%s]" % (place.get_title(), place.gramps_id)
            self.top.get_object("eer_place").set_text(p_lbl)

    def place_delete(self, hndls):
        """Place deleted outside of dialog, remove it if its ours"""
        handle = self.source.get_place_handle()
        if handle and handle in hndls:
            self.source.set_place_handle(None)
            self.top.get_object("eer_place").set_markup(self.place_field.EMPTY_TEXT)
            self.place_field.set_button(False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
# Copyright (C) 2009 Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""File and File format management for the openoffice reports
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from xml.sax.saxutils import escape
import os.path

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docbackend import DocBackend

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".odfbackend.py")

# ------------------------------------------------------------------------
#
# Document Backend class for cairo docs
#
# ------------------------------------------------------------------------


def _escape(string):
    """ a write to the file
    """ """
    change text in text that latex shows correctly
    special characters: & < and >
    """
    string = string.replace("&", "&amp;")  # must be the first
    string = string.replace("<", "&lt;")
    string = string.replace(">", "&gt;")
    return string


class OdfBackend(DocBackend):
    """
    Implementation for open document format docs
    """

    STYLETAG_TO_PROPERTY = {}

    # overwrite base class attributes, they become static var of CairoDoc
    SUPPORTED_MARKUP = [
        DocBackend.BOLD,
        DocBackend.ITALIC,
        DocBackend.UNDERLINE,
        DocBackend.FONTFACE,
        DocBackend.FONTSIZE,
        DocBackend.FONTCOLOR,
        DocBackend.HIGHLIGHT,
        DocBackend.SUPERSCRIPT,
        DocBackend.LINK,
        DocBackend.SUBSCRIPT,
        DocBackend.STRIKETHROUGH,
    ]

    STYLETAG_MARKUP = {
        DocBackend.BOLD: ('<text:span text:style-name="Tbold">', "</text:span>"),
        DocBackend.ITALIC: ('<text:span text:style-name="Titalic">', "</text:span>"),
        DocBackend.UNDERLINE: (
            '<text:span text:style-name="Tunderline">',
            "</text:span>",
        ),
        DocBackend.SUPERSCRIPT: (
            '<text:span text:style-name="GSuper">',
            "</text:span>",
        ),
        DocBackend.SUBSCRIPT: (
            '<text:span text:style-name="GSub">',
            "</text:span>",
        ),
        DocBackend.STRIKETHROUGH: (
            '<text:span text:style-name="GStrikethrough">',
            "</text:span>",
        ),
    }

    ESCAPE_FUNC = lambda x: _escape

    def __init__(self, filename=None):
        """
        @param filename: path name of the file the backend works on
        """
        DocBackend.__init__(self, filename)

    def _create_xmltag(self, tagtype, value):
        """
        overwrites the method in DocBackend
        creates the pango xml tags needed for non bool style types
        """
        if tagtype not in self.SUPPORTED_MARKUP:
            return None
        # The ODF validator does not like spaces or hash in style-names. the
        # font name needs to have the spaces restored, we just hope that the
        # name did not have a hyphen in it originally. The colour is represented
        # without the leading hash, this can be replaced when used in the text-
        # property font colour
        if tagtype == DocBackend.FONTCOLOR:
            return (
                '<text:span text:style-name="FontColor__%s__">'
                % value.replace("#", ""),
                "</text:span>",
            )
        elif tagtype == DocBackend.FONTFACE:
            return (
                '<text:span text:style-name="FontFace__%s__">'
                % self.ESCAPE_FUNC()(value).replace(" ", "-"),
                "</text:span>",
            )
        elif tagtype == DocBackend.FONTSIZE:
            return (
                '<text:span text:style-name="FontSize__%d__">' % value,
                "</text:span>",
            )
        else:  # elif ( tagtype == DocBackend.HIGHLIGHT ):
            return (
                '<text:span text:style-name="FontHighlight__%s__">'
                % value.replace("#", ""),
                "</text:span>",
            )

    def format_link(self, value):
        """
        Override of base method.
        """
        if value.startswith("gramps://"):
            return self.STYLETAG_MARKUP[DocBackend.UNDERLINE]
        else:
            return ('<text:a xlink:href="%s">' % self.ESCAPE_FUNC()(value), "</text:a>")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# gtk
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ... import widgets
from gramps.gen.lib import Event, EventType, Family, FamilyRelType
from .. import build_filter_model
from . import SidebarFilter
from gramps.gen.filters import GenericFilterFactory, rules
from gramps.gen.filters.rules.family import (
    RegExpIdOf,
    RegExpFatherName,
    RegExpMotherName,
    RegExpChildName,
    HasEvent,
    HasRelType,
    HasTag,
    HasNoteRegexp,
    MatchesFilter,
)

GenericFamilyFilter = GenericFilterFactory("Family")


# -------------------------------------------------------------------------
#
# FamilySidebarFilter class
#
# -------------------------------------------------------------------------
class FamilySidebarFilter(SidebarFilter):
    def __init__(self, dbstate, uistate, clicked):
        self.clicked_func = clicked
        self.filter_id = widgets.BasicEntry()
        self.filter_father = widgets.BasicEntry()
        self.filter_mother = widgets.BasicEntry()
        self.filter_child = widgets.BasicEntry()

        self.filter_event = Event()
        self.filter_event.set_type((EventType.CUSTOM, ""))
        self.etype = Gtk.ComboBox(has_entry=True)
        if dbstate.is_open():
            self.custom_types = dbstate.db.get_event_types()
        else:
            self.custom_types = []

        self.event_menu = widgets.MonitoredDataType(
            self.etype,
            self.filter_event.set_type,
            self.filter_event.get_type,
            custom_values=self.custom_types,
        )

        self.filter_family = Family()
        self.filter_family.set_relationship((FamilyRelType.CUSTOM, ""))
        self.rtype = Gtk.ComboBox(has_entry=True)
        if dbstate.is_open():
            self.custom_types = dbstate.db.get_family_relation_types()
        else:
            self.custom_types = []

        self.rel_menu = widgets.MonitoredDataType(
            self.rtype,
            self.filter_family.set_relationship,
            self.filter_family.get_relationship,
            custom_values=self.custom_types,
        )

        self.filter_note = widgets.BasicEntry()

        self.filter_regex = Gtk.CheckButton(label=_("Use regular expressions"))
        self.sensitive_regex = Gtk.CheckButton(label=_("Case sensitive"))

        self.tag = Gtk.ComboBox()
        self.generic = Gtk.ComboBox()

        SidebarFilter.__init__(self, dbstate, uistate, "Family")

    def create_widget(self):
        cell = Gtk.CellRendererText()
        cell.set_property("width", self._FILTER_WIDTH)
        cell.set_property("ellipsize", self._FILTER_ELLIPSIZE)
        self.generic.pack_start(cell, True)
        self.generic.add_attribute(cell, "text", 0)
        self.on_filters_changed("Family")

        cell = Gtk.CellRendererText()
        cell.set_property("width", self._FILTER_WIDTH)
        cell.set_property("ellipsize", self._FILTER_ELLIPSIZE)
        self.tag.pack_start(cell, True)
        self.tag.add_attribute(cell, "text", 0)

        self.etype.get_child().set_width_chars(5)
        self.rtype.get_child().set_width_chars(5)

        self.add_text_entry(_("ID"), self.filter_id)
        self.add_text_entry(_("Father"), self.filter_father)
        self.add_text_entry(_("Mother"), self.filter_mother)
        self.add_text_entry(_("Child"), self.filter_child)
        self.add_entry(_("Relationship"), self.rtype)
        self.add_entry(_("Family Event"), self.etype)
        self.add_text_entry(_("Family Note"), self.filter_note)
        self.add_entry(_("Tag"), self.tag)
        self.add_filter_entry(_("Custom filter"), self.generic)
        self.add_regex_entry(self.filter_regex)
        self.add_regex_case(self.sensitive_regex)

    def clear(self, obj):
        self.filter_id.set_text("")
        self.filter_father.set_text("")
        self.filter_mother.set_text("")
        self.filter_child.set_text("")
        self.filter_note.set_text("")
        self.etype.get_child().set_text("")
        self.rtype.get_child().set_text("")
        self.tag.set_active(0)
        self.generic.set_active(0)

    def get_filter(self):
        gid = str(self.filter_id.get_text()).strip()
        father = str(self.filter_father.get_text()).strip()
        mother = str(self.filter_mother.get_text()).strip()
        child = str(self.filter_child.get_text()).strip()
        note = str(self.filter_note.get_text()).strip()
        etype = self.filter_event.get_type().xml_str()
        rtype = self.filter_family.get_relationship().xml_str()
        regex = self.filter_regex.get_active()
        usecase = self.sensitive_regex.get_active()
        tag = self.tag.get_active() > 0
        generic = self.generic.get_active() > 0

        empty = not (
            gid
            or father
            or mother
            or child
            or note
            or regex
            or etype
            or rtype
            or tag
            or generic
        )
        if empty:
            generic_filter = None
        else:
            generic_filter = GenericFamilyFilter()
            if gid:
                rule = RegExpIdOf([gid], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            if father:
                # if the name is not empty, split the name in multiple part if
                # we don't use regexp. if the regexp is used, don't split the
                # field
                if not regex:
                    name_parts = father.split(sep=" ")
                    for name_part in name_parts:
                        rule = RegExpFatherName(
                            [name_part], use_regex=regex, use_case=usecase
                        )
                        generic_filter.add_rule(rule)
                else:
                    rule = RegExpFatherName([father], use_regex=regex, use_case=usecase)
                    generic_filter.add_rule(rule)

            if mother:
                # if the name is not empty, split the name in multiple part if
                # we don't use regexp. if the regexp is used, don't split the
                # field
                if not regex:
                    name_parts = mother.split(sep=" ")
                    for name_part in name_parts:
                        rule = RegExpMotherName(
                            [name_part], use_regex=regex, use_case=usecase
                        )
                        generic_filter.add_rule(rule)
                else:
                    rule = RegExpMotherName([mother], use_regex=regex, use_case=usecase)
                    generic_filter.add_rule(rule)

            if child:
                # if the name is not empty, split the name in multiple part if
                # we don't use regexp. if the regexp is used, don't split the
                # field
                if not regex:
                    name_parts = child.split(sep=" ")
                    for name_part in name_parts:
                        rule = RegExpChildName(
                            [name_part], use_regex=regex, use_case=usecase
                        )
                        generic_filter.add_rule(rule)
                else:
                    rule = RegExpChildName([child], use_regex=regex, use_case=usecase)
                    generic_filter.add_rule(rule)

            if etype:
                rule = HasEvent(
                    [etype, "", "", "", ""], use_regex=regex, use_case=usecase
                )
                generic_filter.add_rule(rule)

            if rtype:
                rule = HasRelType([rtype], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            if note:
                rule = HasNoteRegexp([note], use_regex=regex, use_case=usecase)
                generic_filter.add_rule(rule)

            # check the Tag
            if tag:
                model = self.tag.get_model()
                node = self.tag.get_active_iter()
                attr = model.get_value(node, 0)
                rule = HasTag([attr])
                generic_filter.add_rule(rule)

        if self.generic.get_active() != 0:
            model = self.generic.get_model()
            node = self.generic.get_active_iter()
            obj = str(model.get_value(node, 0))
            rule = MatchesFilter([obj])
            generic_filter.add_rule(rule)

        return generic_filter

    def on_filters_changed(self, name_space):
        if name_space == "Family":
            all_filter = GenericFamilyFilter()
            all_filter.set_name(_("None"))
            all_filter.add_rule(rules.family.AllFamilies([]))
            self.generic.set_model(build_filter_model("Family", [all_filter]))
            self.generic.set_active(0)

    def on_tags_changed(self, tag_list):
        """
        Update the list of tags in the tag filter.
        """
        model = Gtk.ListStore(str)
        model.append(("",))
        for tag_name in tag_list:
            model.append((tag_name,))
        self.tag.set_model(model)
        self.tag.set_active(0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Database Processing/Extract Event Descriptions from Event Data"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# gnome/gtk
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gen.lib import EventRoleType
from gramps.gen.db import DbTxn
from gramps.gen.utils.db import family_name

from gramps.gui.plug import tool
from gramps.gen.display.name import displayer as name_displayer


# -------------------------------------------------------------------------
#
# EventNames
#
# -------------------------------------------------------------------------
class EventNames(tool.BatchTool):
    """
    Look for events that do not have a description, and build the description
    from the item that contains it.

    Looks for a PRIMARY role type for events attached to a persons, and a
    FAMILY role for an event that is attached to a family.

    """

    def __init__(self, dbstate, user, options_class, name, callback=None):
        self.user = user
        tool.BatchTool.__init__(self, dbstate, user, options_class, name)

        if not self.fail:
            self.run()

    def run(self):
        """
        Perform the actual extraction of information.
        """
        with DbTxn(_("Event name changes"), self.db, batch=True) as trans:
            self.db.disable_signals()
            self.change = False
            counter = 0

            with self.user.progress(_("Extract Event Description"), "", 2) as step:
                for person in self.db.iter_people():
                    for event_ref in person.get_event_ref_list():
                        if event_ref.get_role() == EventRoleType.PRIMARY:
                            event_handle = event_ref.ref
                            event = self.db.get_event_from_handle(event_handle)
                            if event.get_description() == "":
                                person_event_name(event, person)
                                self.db.commit_event(event, trans)
                                self.change = True
                                counter += 1
                step()

                for family in self.db.iter_families():
                    for event_ref in family.get_event_ref_list():
                        if event_ref.get_role() == EventRoleType.FAMILY:
                            event_handle = event_ref.ref
                            event = self.db.get_event_from_handle(event_handle)
                            if event.get_description() == "":
                                family_event_name(event, family, self.db)
                                self.db.commit_event(event, trans)
                                self.change = True
                                counter += 1
                step()

        self.db.enable_signals()
        self.db.request_rebuild()

        if hasattr(self.user.uistate, "window"):
            parent_window = self.user.uistate.window
        else:
            parent_window = None
        if self.change == True:
            # Translators: leave all/any {...} untranslated
            message = ngettext(
                "{quantity} event description has been added",
                "{quantity} event descriptions have been added",
                counter,
            ).format(quantity=counter)
            self.user.info(_("Modifications made"), message, parent=parent_window)
        else:
            self.user.info(
                _("No modifications made"),
                _("No event description has been added."),
                parent=parent_window,
            )


# -------------------------------------------------------------------------
#
# Support functions
#
# -------------------------------------------------------------------------

# feature requests 2356, 1658: avoid genitive form
EVENT_FAMILY_STR = _("%(event_name)s of %(family)s")
# feature requests 2356, 1658: avoid genitive form
EVENT_PERSON_STR = _("%(event_name)s of %(person)s")


def person_event_name(event, person):
    """
    Build a name for an event based on the primary person's information.
    """
    if not event.get_description():
        text = EVENT_PERSON_STR % {
            "event_name": str(event.get_type()),
            "person": name_displayer.display(person),
        }
        event.set_description(text)


def family_event_name(event, family, dbase):
    """
    Build a name for an event based on the family's information.
    """
    if not event.get_description():
        text = EVENT_FAMILY_STR % {
            "event_name": str(event.get_type()),
            "family": family_name(family, dbase),
        }
        event.set_description(text)


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class EventNamesOptions(tool.ToolOptions):
    """
    Define options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Handling of loading new/existing databases.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import logging

# -------------------------------------------------------------------------
#
# Set up logging
#
# -------------------------------------------------------------------------
_LOG = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# GTK+ modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.db.dbconst import DBBACKEND
from gramps.gen.db.utils import make_database
from gramps.gen.db.upgrade import make_zip_backup

_ = glocale.translation.gettext
from gramps.cli.grampscli import CLIDbLoader
from gramps.gen.config import config
from gramps.gen.db.exceptions import (
    DbUpgradeRequiredError,
    DbVersionError,
    DbPythonError,
    DbSupportedError,
    DbConnectionError,
)
from .pluginmanager import GuiPluginManager
from .dialog import DBErrorDialog, ErrorDialog, QuestionDialog2, WarningDialog
from .user import User
from gramps.gen.errors import DbError
from .managedwindow import ManagedWindow


# -------------------------------------------------------------------------
#
# DbLoader class
#
# -------------------------------------------------------------------------
class DbLoader(CLIDbLoader):
    def __init__(self, dbstate, uistate):
        CLIDbLoader.__init__(self, dbstate)
        self.uistate = uistate
        self.import_info = None

    def _warn(self, title, warnmessage):
        WarningDialog(title, warnmessage, parent=self.uistate.window)

    def _errordialog(self, title, errormessage):
        """
        Show the error.
        In the GUI, the error is shown, and a return happens
        """
        ErrorDialog(title, errormessage, parent=self.uistate.window)
        return 1

    def _dberrordialog(self, msg):
        import traceback

        exc = traceback.format_exc()
        try:
            DBErrorDialog(str(msg.value), parent=self.uistate.window)
            _LOG.error(str(msg.value))
        except:
            DBErrorDialog(str(msg), parent=self.uistate.window)
            _LOG.error(str(msg) + "\n" + exc)

    def import_file(self):
        self.import_info = None
        # First thing first: import is a batch transaction
        # so we will lose the undo history. Warn the user.

        if self.dbstate.db.get_number_of_people() > 0:
            warn_dialog = QuestionDialog2(
                _("Undo history warning"),
                _(
                    "Proceeding with import will erase the undo history "
                    "for this session. In particular, you will not be able "
                    "to revert the import or any changes made prior to it.\n\n"
                    "If you think you may want to revert the import, "
                    "please stop here and backup your database."
                ),
                _("_Proceed with import"),
                _("_Stop"),
                parent=self.uistate.window,
            )
            if not warn_dialog.run():
                return False

        GrampsImportFileDialog(self.dbstate, self.uistate, callback=self.set_info)

    def set_info(self, info):
        self.import_info = info

    def import_info_text(self):
        """
        On import the importer can construct an info object about the import.
        If so, this method will return this text, otherwise the empty string
        is returned
        """
        if self.import_info is None:
            return ""
        return self.import_info.info_text()

    def read_file(self, filename, username=None, password=None):
        """
        This method takes care of changing database, and loading the data.
        In 3.0 we only allow reading of real databases of filetype
        'x-directory/normal'

        This method should only return on success.
        Returning on failure makes no sense, because we cannot recover,
        since database has already been changed.
        Therefore, any errors should raise exceptions.

        On success, return with the disabled signals. The post-load routine
        should enable signals, as well as finish up with other UI goodies.
        """

        if os.path.exists(filename):
            if not os.access(filename, os.W_OK):
                mode = "r"
                self._warn(
                    _("Read only database"),
                    _("You do not have write access " "to the selected file."),
                )
            else:
                mode = "w"
        else:
            mode = "w"

        self.dbstate.no_database()

        self.uistate.progress.show()
        self.uistate.pulse_progressbar(0)

        force_schema_upgrade = False
        try:
            while True:
                dbid_path = os.path.join(filename, DBBACKEND)
                if os.path.isfile(dbid_path):
                    with open(dbid_path) as fp:
                        dbid = fp.read().strip()
                else:
                    dbid = "bsddb"

                db = make_database(dbid)
                db.disable_signals()
                if db.requires_login() and username is None:
                    login = GrampsLoginDialog(self.uistate)
                    credentials = login.run()
                    if credentials is None:
                        return
                    username, password = credentials

                try:
                    db.load(
                        filename,
                        self.uistate.pulse_progressbar,
                        mode,
                        force_schema_upgrade,
                        username=username,
                        password=password,
                    )
                    if self.dbstate.is_open():
                        self.dbstate.db.close(
                            user=User(
                                callback=self.uistate.pulse_progressbar,
                                uistate=self.uistate,
                                dbstate=self.dbstate,
                            )
                        )
                    self.dbstate.change_database(db)
                    break
                except (DbSupportedError, DbUpgradeRequiredError) as msg:
                    if (
                        force_schema_upgrade
                        or QuestionDialog2(
                            _("Are you sure you want " "to upgrade this Family Tree?"),
                            str(msg),
                            _(
                                "I have made a backup,\n"
                                "please upgrade my Family Tree"
                            ),
                            _("Cancel"),
                            parent=self.uistate.window,
                        ).run()
                    ):
                        force_schema_upgrade = True
                        make_zip_backup(filename)
                    else:
                        self.dbstate.no_database()
                        break
        # Get here is there is an exception the while loop does not handle
        except DbVersionError as msg:
            self.dbstate.no_database()
            self._errordialog(_("Cannot open database"), str(msg))
        except DbPythonError as msg:
            self.dbstate.no_database()
            self._errordialog(_("Cannot open database"), str(msg))
        except DbConnectionError as msg:
            self.dbstate.no_database()
            self._warn(_("Cannot open database"), str(msg))
        except OSError as msg:
            self.dbstate.no_database()
            self._errordialog(_("Could not open file: %s") % filename, str(msg))
        except DbError as msg:
            self.dbstate.no_database()
            self._dberrordialog(msg)
        except Exception as newerror:
            self.dbstate.no_database()
            self._dberrordialog(str(newerror))

        self.uistate.progress.hide()
        return True


# -------------------------------------------------------------------------
#
# FileChooser filters: what to show in the file chooser
#
# -------------------------------------------------------------------------
def add_all_files_filter(chooser):
    """
    Add an all-permitting filter to the file chooser dialog.
    """
    mime_filter = Gtk.FileFilter()
    mime_filter.set_name(_("All files"))
    mime_filter.add_pattern("*")
    chooser.add_filter(mime_filter)


def icase(ext):
    """
    Return a glob reresenting a case insensitive file extension.
    """
    return "".join(["[{}{}]".format(s.lower(), s.upper()) for s in ext])


# -------------------------------------------------------------------------
#
# Format selectors: explictly set the format of the file
#
# -------------------------------------------------------------------------
class GrampsFormatWidget(Gtk.ComboBox):
    def __init__(self):
        Gtk.ComboBox.__init__(self, model=None)

    def set(self, format_list):
        self.store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(self.store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)
        self.format_list = format_list

        for format, label in format_list:
            self.store.append(row=[label])
        self.set_active(False)

    def get_value(self):
        active = self.get_active()
        if active < 0:
            return None
        return self.format_list[active][0]


def format_maker():
    """
    A factory function making format selection widgets.

    Accepts a list of formats to include into selector.
    The auto selection is always added as the first one.
    The returned box contains both the label and the selector.
    """
    pmgr = GuiPluginManager.get_instance()
    format_list = [("auto", _("Automatically detected"))]

    for plugin in pmgr.get_import_plugins():
        format_list.append((plugin.get_extension(), plugin.get_name()))

    type_selector = GrampsFormatWidget()
    type_selector.set(format_list)

    box = Gtk.Box()
    label = Gtk.Label(label=_("Select file _type:"))
    label.set_use_underline(True)
    label.set_mnemonic_widget(type_selector)
    box.pack_start(label, expand=False, fill=False, padding=6)
    box.add(type_selector)
    box.show_all()
    return (box, type_selector)


class GrampsLoginDialog(ManagedWindow):
    def __init__(self, uistate):
        """
        A login dialog to obtain credentials to connect to a database
        """
        self.title = _("Login")
        ManagedWindow.__init__(self, uistate, [], self.__class__, modal=True)

        dialog = Gtk.Dialog(transient_for=uistate.window)
        grid = Gtk.Grid()
        grid.set_border_width(6)
        grid.set_row_spacing(6)
        grid.set_column_spacing(6)
        label = Gtk.Label(label=_("Username: "))
        grid.attach(label, 0, 0, 1, 1)
        self.username = Gtk.Entry()
        self.username.set_hexpand(True)
        grid.attach(self.username, 1, 0, 1, 1)
        label = Gtk.Label(label=_("Password: "))
        grid.attach(label, 0, 1, 1, 1)
        self.password = Gtk.Entry()
        self.password.set_hexpand(True)
        self.password.set_visibility(False)
        self.password.set_input_purpose(Gtk.InputPurpose.PASSWORD)
        grid.attach(self.password, 1, 1, 1, 1)
        dialog.vbox.pack_start(grid, True, True, 0)
        dialog.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("Login"), Gtk.ResponseType.OK
        )
        self.set_window(dialog, None, self.title)

    def run(self):
        self.show()
        response = self.window.run()
        username = self.username.get_text()
        password = self.password.get_text()
        if response == Gtk.ResponseType.CANCEL:
            self.close()
            return None
        elif response == Gtk.ResponseType.OK:
            self.close()
            return (username, password)


class GrampsImportFileDialog(ManagedWindow):
    def __init__(self, dbstate, uistate, callback=None):
        """
        A dialog to import a file into Gramps
        """
        self.dbstate = dbstate

        self.title = _("Import Family Tree")
        ManagedWindow.__init__(self, uistate, [], self.__class__, modal=True)
        # the import_dialog.run() below makes it modal, so any change to
        # the previous line's "modal" would require that line to be changed

        pmgr = GuiPluginManager.get_instance()

        import_dialog = Gtk.FileChooserDialog(
            title="",
            transient_for=self.uistate.window,
            action=Gtk.FileChooserAction.OPEN,
        )
        import_dialog.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("Import"), Gtk.ResponseType.OK
        )
        self.set_window(import_dialog, None, self.title)
        self.setup_configs("interface.grampsimportfiledialog", 780, 630)
        import_dialog.set_local_only(False)

        # Add all supported files depending on available plugins
        gramps_filter = Gtk.FileFilter()
        gramps_filter.set_name(_("All supported files"))
        import_dialog.add_filter(gramps_filter)

        # Always add automatic (match all files) filter
        add_all_files_filter(import_dialog)  # *

        # Add more file type selections for available importers
        for plugin in pmgr.get_import_plugins():
            file_filter = Gtk.FileFilter()
            name = "%s (.%s)" % (plugin.get_name(), plugin.get_extension())
            file_filter.set_name(name)
            file_filter.add_pattern("*.%s" % icase(plugin.get_extension()))
            gramps_filter.add_pattern("*.%s" % icase(plugin.get_extension()))
            import_dialog.add_filter(file_filter)

        (box, type_selector) = format_maker()
        import_dialog.set_extra_widget(box)

        import_dialog.set_current_folder(config.get("paths.recent-import-dir"))
        while True:
            # the import_dialog.run() makes it modal, so any change to that
            # line would require the ManagedWindow.__init__ to be changed also
            response = import_dialog.run()
            if response == Gtk.ResponseType.CANCEL:
                break
            elif response == Gtk.ResponseType.DELETE_EVENT:
                return
            elif response == Gtk.ResponseType.OK:
                filename = import_dialog.get_filename()
                if self.check_errors(filename):
                    # displays errors if any
                    continue

                (the_path, the_file) = os.path.split(filename)
                config.set("paths.recent-import-dir", the_path)

                extension = type_selector.get_value()
                if extension == "auto":
                    # Guess the file format based on the file extension.
                    # This will get the lower case extension without a period,
                    # or an empty string.
                    extension = os.path.splitext(filename)[-1][1:].lower()

                for plugin in pmgr.get_import_plugins():
                    if extension == plugin.get_extension():
                        self.close()
                        self.do_import(plugin.get_import_function(), filename)
                        if callback is not None:
                            callback(self.import_info)
                        return

                # Finally, we give up and declare this an unknown format
                ErrorDialog(
                    _("Could not open file: %s") % filename,
                    _(
                        'File type "%s" is unknown to Gramps.\n\n'
                        "Valid types are: Gramps database, Gramps XML, "
                        "Gramps package, GEDCOM, and others."
                    )
                    % extension,
                    parent=self.uistate.window,
                )

        self.close()

    def check_errors(self, filename):
        """
        Run common error checks and return True if any found.

        In this process, a warning dialog can pop up.

        """
        if not isinstance(filename, str):
            return True

        filename = os.path.normpath(os.path.abspath(filename))

        if len(filename) == 0:
            return True
        elif os.path.isdir(filename):
            ErrorDialog(
                _("Cannot open file"),
                _("The selected file is a directory, not a file.\n"),
                parent=self.uistate.window,
            )
            return True
        elif os.path.exists(filename):
            if not os.access(filename, os.R_OK):
                ErrorDialog(
                    _("Cannot open file"),
                    _("You do not have read access to the selected file."),
                    parent=self.uistate.window,
                )
                return True
        else:
            try:
                f = open(filename, "w")
                f.close()
                os.remove(filename)
            except IOError:
                ErrorDialog(
                    _("Cannot create file"),
                    _("You do not have write access to the selected file."),
                    parent=self.uistate.window,
                )
                return True

        return False

    def do_import(self, importer, filename):
        self.import_info = None
        self._begin_progress()
        self.uistate.set_sensitive(False)

        try:
            # an importer can return an object with info, object.info_text()
            # returns that info. Otherwise None is set to import_info
            self.import_info = importer(
                self.dbstate.db,
                filename,
                User(
                    callback=self._pulse_progress,
                    uistate=self.uistate,
                    dbstate=self.dbstate,
                ),
            )
            dirname = os.path.dirname(filename) + os.path.sep
            config.set("paths.recent-import-dir", dirname)
        except UnicodeError as msg:
            ErrorDialog(
                _("Could not import file: %s") % filename,
                _(
                    "This file incorrectly identifies its character "
                    "set, so it cannot be accurately imported. Please fix the "
                    "encoding, and import again"
                )
                + "\n\n %s" % msg,
                parent=self.uistate.window,
            )
        except Exception:
            _LOG.error("Failed to import database.", exc_info=True)
        self.uistate.set_sensitive(True)
        self._end_progress()

    def build_menu_names(self, obj):  # this is meaningless since it's modal
        return (self.title, None)

    def _begin_progress(self):
        self.uistate.set_busy_cursor(True)
        self.uistate.progress.show()
        self.uistate.pulse_progressbar(0)

    def _pulse_progress(self, value):
        self.uistate.pulse_progressbar(value)

    def _end_progress(self):
        self.uistate.set_busy_cursor(False)
        self.uistate.progress.hide()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2011 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
GrampletView interface.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango
from xml.sax.saxutils import escape
import time
import os
import configparser

import logging

LOG = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.errors import WindowActiveError
from gramps.gen.const import URL_MANUAL_PAGE, VERSION_DIR, COLON
from gramps.gen.utils.configmanager import clean_up
from ..editors import EditPerson, EditFamily
from ..managedwindow import ManagedWindow
from ..utils import is_right_click, match_primary_mask, get_link_color
from ..uimanager import ActionGroup, valid_action_name
from ..plug import make_gui_option
from ..plug.quick import run_quick_report_by_name
from ..display import display_help, display_url
from ..glade import Glade
from ..pluginmanager import GuiPluginManager
from .undoablebuffer import UndoableBuffer
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = URL_MANUAL_PAGE + "_-_Gramplets"

# -------------------------------------------------------------------------
#
# Globals
#
# -------------------------------------------------------------------------
PLUGMAN = GuiPluginManager.get_instance()
NL = "\n"


def AVAILABLE_GRAMPLETS():
    return [gplug.id for gplug in PLUGMAN.get_reg_gramplets()]


def GET_AVAILABLE_GRAMPLETS(name):
    for gplug in PLUGMAN.get_reg_gramplets():
        if gplug.id == name:
            return {
                "name": gplug.id,
                "tname": gplug.name,
                "version": gplug.version,
                "height": gplug.height,
                "expand": gplug.expand,
                "title": gplug.gramplet_title,  # translated
                "content": gplug.gramplet,
                "detached_width": gplug.detached_width,
                "detached_height": gplug.detached_height,
                "state": "maximized",
                "gramps": "0.0.0",
                "column": -1,
                "row": -1,
                "page": 0,
                "data": [],
                "help_url": gplug.help_url,
                "navtypes": gplug.navtypes,
            }
    return None


def GET_GRAMPLET_LIST(nav_type, skip):
    return [
        (gplug.gramplet_title, gplug.id)
        for gplug in PLUGMAN.get_reg_gramplets()
        if (gplug.navtypes == [] or nav_type in gplug.navtypes)
        and gplug.name not in skip
    ]


def parse_tag_attr(text):
    """
    Function used to parse markup.
    """
    text = text.strip()
    parts = text.split(" ", 1)
    attrs = {}
    if len(parts) == 2:
        attr_values = parts[1].split(" ")  # "name=value name=value"
        for av in attr_values:
            attribute, value = av.split("=", 1)
            value = value.strip()
            # trim off quotes:
            if value[0] == value[-1] and value[0] in ['"', "'"]:
                value = value[1:-1]
            attrs[attribute.strip().lower()] = value
    return [parts[0].upper(), attrs]


def get_gramplet_opts(name, opts):
    """
    Lookup the options for a given gramplet name and update
    the options with provided dictionary, opts.
    """
    if name in AVAILABLE_GRAMPLETS():
        data = GET_AVAILABLE_GRAMPLETS(name)
        my_data = data.copy()
        my_data.update(opts)
        return my_data
    else:
        LOG.warning("Unknown gramplet name: '%s'", name)
        return {}


def get_gramplet_options_by_name(name):
    """
    Get options by gramplet name.
    """
    if name in AVAILABLE_GRAMPLETS():
        return GET_AVAILABLE_GRAMPLETS(name).copy()
    else:
        LOG.warning("Unknown gramplet name: '%s'", name)
        return None


def get_gramplet_options_by_tname(name):
    """
    get options by translated name.
    """
    for key in AVAILABLE_GRAMPLETS():
        if GET_AVAILABLE_GRAMPLETS(key)["tname"] == name:
            return GET_AVAILABLE_GRAMPLETS(key).copy()
    LOG.warning("Unknown gramplet name: '%s'", name)
    return None


def make_requested_gramplet(gui_class, pane, opts, dbstate, uistate):
    """
    Make a GUI gramplet given its name.
    """
    if opts is None:
        return None

    if "name" in opts:
        name = opts["name"]
        if name in AVAILABLE_GRAMPLETS():
            gui = gui_class(pane, dbstate, uistate, **opts)
            if opts.get("content", None):
                pdata = PLUGMAN.get_plugin(name)
                module = PLUGMAN.load_plugin(pdata)
                if module:
                    getattr(module, opts["content"])(gui)
                else:
                    LOG.warning(
                        "Error loading gramplet '%s': " "skipping content", name
                    )
            return gui
    else:
        LOG.warning("Error loading gramplet: unknown name")
    return None


def logical_true(value):
    """
    Used for converting text file values to booleans.
    """
    return value in ["True", True, 1, "1"]


def make_callback(func, arg):
    """
    Generates a callback function based off the passed arguments
    """
    return lambda x, y: func(arg)


class LinkTag(Gtk.TextTag):
    """
    Class for keeping track of link data.
    """

    lid = 0
    # obtaining the theme link color once. Restart needed on theme change!
    linkcolor = Gtk.Label(label="test")  # needed to avoid label destroyed to early
    linkcolor = get_link_color(linkcolor.get_style_context())

    def __init__(self, buffer):
        LinkTag.lid += 1
        Gtk.TextTag.__init__(self, name=str(LinkTag.lid))
        tag_table = buffer.get_tag_table()
        self.set_property("foreground", self.linkcolor)
        # self.set_property('underline', Pango.Underline.SINGLE)
        try:
            tag_table.add(self)
        except ValueError:  # tag is already in tag table
            pass


class GrampletWindow(ManagedWindow):
    """
    Class for showing a detached gramplet.
    """

    def __init__(self, gramplet):
        """
        Constructs the window, and loads the GUI gramplet.
        """
        self.title = gramplet.title + " " + _("Gramplet")
        self.gramplet = gramplet
        self.gramplet.scrolledwindow.set_vexpand(True)
        self.gramplet.detached_window = self
        # Keep track of what state it was in:
        self.docked_state = gramplet.gstate
        # Now detach it
        self.gramplet.set_state("detached")
        ManagedWindow.__init__(self, gramplet.uistate, [], self.title)
        dlg = Gtk.Dialog(
            transient_for=gramplet.uistate.window, destroy_with_parent=True
        )
        dlg.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_window(dlg, None, self.title)
        cfg_name = gramplet.gname.replace(" ", "").lower() + "-gramplet"
        self.setup_configs(
            "interface." + cfg_name, gramplet.detached_width, gramplet.detached_height
        )
        self.window.add_button(_("_Help"), Gtk.ResponseType.HELP)
        # add gramplet:
        if self.gramplet.pui:
            self.gramplet.pui.active = True
        self.gramplet.mainframe.get_parent().remove(self.gramplet.mainframe)
        self.window.vbox.add(self.gramplet.mainframe)
        self.window.connect("response", self.handle_response)
        self.show()
        # After we show, then we hide:
        self.gramplet.gvclose.hide()
        self.gramplet.gvstate.hide()
        self.gramplet.gvproperties.hide()
        if self.gramplet.titlelabel_entry:
            self.gramplet.titlelabel_entry.hide()
        if self.gramplet.pui:
            for widget in self.gramplet.pui.hidden_widgets():
                widget.hide()

    def handle_response(self, object, response):
        """
        Callback for taking care of button clicks.
        """
        if response == Gtk.ResponseType.CLOSE:
            self.close()
        elif response == Gtk.ResponseType.HELP:
            # translated name:
            if self.gramplet.help_url:
                if self.gramplet.help_url.startswith(("http://", "https://")):
                    display_url(self.gramplet.help_url)
                else:
                    display_help(self.gramplet.help_url)
            else:
                display_help(WIKI_HELP_PAGE, self.gramplet.tname.replace(" ", "_"))

    def build_menu_names(self, obj):
        """
        Part of the Gramps window interface.
        """
        return (self.title, "Gramplet")

    def get_title(self):
        """
        Returns the window title.
        """
        return self.title

    def close(self, *args):
        """
        Dock the detached GrampletWindow back in the column from where it came.
        """
        self.gramplet.scrolledwindow.set_vexpand(False)
        self.gramplet.detached_window = None
        self.gramplet.pane.detached_gramplets.remove(self.gramplet)
        if self.docked_state == "minimized":
            self.gramplet.set_state("minimized")
        else:
            self.gramplet.set_state("maximized")
        pane = self.gramplet.pane
        col = self.gramplet.column
        stack = []
        for gframe in pane.columns[col]:
            gramplet = pane.frame_map[str(gframe)]
            if gramplet.row > self.gramplet.row:
                pane.columns[col].remove(gframe)
                stack.append(gframe)
        expand = self.gramplet.gstate == "maximized" and self.gramplet.expand
        column = pane.columns[col]
        parent = self.gramplet.pane.get_column_frame(self.gramplet.column)
        self.gramplet.mainframe.get_parent().remove(self.gramplet.mainframe)
        parent.add(self.gramplet.mainframe)
        if self.gramplet.pui:
            self.gramplet.pui.active = self.gramplet.pane.pageview.active
        for gframe in stack:
            gramplet = pane.frame_map[str(gframe)]
            expand = gramplet.gstate == "maximized" and gramplet.expand
            pane.columns[col].pack_start(gframe, expand, True, 0)
        # Now make sure they all have the correct expand:
        for gframe in pane.columns[col]:
            gramplet = pane.frame_map[str(gframe)]
            expand, fill, padding, pack = column.query_child_packing(gramplet.mainframe)
            expand = gramplet.gstate == "maximized" and gramplet.expand
            column.set_child_packing(gramplet.mainframe, expand, fill, padding, pack)
        # set_image on buttons as get_image is None in first run
        # or point to invalid adress in every other run
        self.gramplet.gvstate.set_image(self.gramplet.xml.get_object("gvstateimage"))
        self.gramplet.gvclose.set_image(self.gramplet.xml.get_object("gvcloseimage"))
        self.gramplet.gvproperties.set_image(
            self.gramplet.xml.get_object("gvpropertiesimage")
        )
        self.gramplet.gvclose.show()
        self.gramplet.gvstate.show()
        self.gramplet.gvproperties.show()
        ManagedWindow.close(self, *args)


# ------------------------------------------------------------------------


class GuiGramplet:
    """
    Class that handles the GUI representation of a Gramplet.
    """

    def __init__(self, pane, dbstate, uistate, title, **kwargs):
        """
        Internal constructor for GUI portion of a gramplet.
        """
        self.pane = pane
        self.view = pane.pageview
        self.dbstate = dbstate
        self.uistate = uistate
        self.track = []
        self.title = title
        self.detached_window = None
        self.force_update = False
        self.title_override = False
        self._tags = []
        ########## Set defaults
        self.gname = kwargs.get("name", "Unnamed Gramplet")
        self.tname = kwargs.get("tname", "Unnamed Gramplet")
        self.navtypes = kwargs.get("navtypes", [])
        self.version = kwargs.get("version", "0.0.0")
        self.gramps = kwargs.get("gramps", "0.0.0")
        self.expand = logical_true(kwargs.get("expand", False))
        self.height = int(kwargs.get("height", 200))
        self.width = int(kwargs.get("width", 375))
        self.column = int(kwargs.get("column", -1))
        self.detached_height = int(kwargs.get("detached_height", 300))
        self.detached_width = int(kwargs.get("detached_width", 400))
        self.row = int(kwargs.get("row", -1))
        self.page = int(kwargs.get("page", -1))
        self.gstate = kwargs.get("state", "maximized")
        self.data = kwargs.get("data", [])
        self.help_url = kwargs.get("help_url", WIKI_HELP_PAGE)
        if self.help_url == "None":
            self.help_url = None  # to fix up the config file vers of None
        ##########
        self.use_markup = False
        self.pui = None  # user code
        self.tooltips_text = None

        self.link_cursor = Gdk.Cursor.new_for_display(
            Gdk.Display.get_default(), Gdk.CursorType.LEFT_PTR
        )
        self.standard_cursor = Gdk.Cursor.new_for_display(
            Gdk.Display.get_default(), Gdk.CursorType.XTERM
        )

        self.scrolledwindow = None
        self.textview = None
        self.buffer = None

    def set_tooltip(self, tip):
        self.tooltips_text = tip
        self.scrolledwindow.set_tooltip_text(tip)

    def undo(self):
        self.buffer.undo()
        self.text_length = len(self.get_text())

    def redo(self):
        self.buffer.redo()
        self.text_length = len(self.get_text())

    def on_key_press_event(self, widget, event):
        """Signal handler.

        Handle formatting shortcuts.

        """
        if (Gdk.keyval_name(event.keyval) == "Z") and match_primary_mask(
            event.get_state(), Gdk.ModifierType.SHIFT_MASK
        ):
            self.redo()
            return True
        elif (Gdk.keyval_name(event.keyval) == "z") and match_primary_mask(
            event.get_state()
        ):
            self.undo()
            return True

        return False

    def append_text(self, text, scroll_to="end"):
        enditer = self.buffer.get_end_iter()
        start = self.buffer.create_mark(None, enditer, True)
        self.buffer.insert(enditer, text)
        self.text_length += len(text)
        if scroll_to == "end":
            enditer = self.buffer.get_end_iter()
            end = self.buffer.create_mark(None, enditer, True)
            self.textview.scroll_to_mark(end, 0.0, True, 0, 0)
        elif scroll_to == "start":  # beginning of this append
            self.textview.scroll_to_mark(start, 0.0, True, 0, 0)
        elif scroll_to == "begin":  # beginning of the buffer
            begin_iter = self.buffer.get_start_iter()
            begin = self.buffer.create_mark(None, begin_iter, True)
            self.textview.scroll_to_mark(begin, 0.0, True, 0, 0)
        else:
            raise AttributeError("no such cursor position: '%s'" % scroll_to)

    def clear_text(self):
        self.buffer.set_text("")
        self.text_length = 0

    def get_text(self):
        start = self.buffer.get_start_iter()
        end = self.buffer.get_end_iter()
        return self.buffer.get_text(start, end, True)  # include invisible chars

    def insert_text(self, text):
        self.buffer.insert_at_cursor(text)
        self.text_length += len(text)

    def render_text(self, text):
        markup_pos = {"B": [], "I": [], "U": [], "A": [], "TT": []}
        retval = ""
        i = 0
        r = 0
        tag = ""
        while i < len(text):
            if text[i : i + 2] == "</":
                # start of ending tag
                stop = text[i:].find(">")
                if stop < 0:
                    retval += text[i]
                    r += 1
                    i += 1
                else:
                    markup = text[i + 2 : i + stop].upper()  # close tag
                    markup_pos[markup][-1].append(r)
                    i += stop + 1
            elif text[i] == "<":
                # start of start tag
                stop = text[i:].find(">")
                if stop < 0:
                    retval += text[i]
                    r += 1
                    i += 1
                else:
                    markup, attr = parse_tag_attr(text[i + 1 : i + stop])
                    markup_pos[markup].append([r, attr])
                    i += stop + 1
            elif text[i] == "\\":
                retval += text[i + 1]
                r += 1
                i += 2
            else:
                retval += text[i]
                r += 1
                i += 1
        offset = self.text_length
        self.append_text(retval)
        for items in markup_pos["TT"]:
            if len(items) == 3:
                (a, attributes, b) = items
                start = self.buffer.get_iter_at_offset(a + offset)
                stop = self.buffer.get_iter_at_offset(b + offset)
                self.buffer.apply_tag_by_name("fixed", start, stop)
        for items in markup_pos["B"]:
            if len(items) == 3:
                (a, attributes, b) = items
                start = self.buffer.get_iter_at_offset(a + offset)
                stop = self.buffer.get_iter_at_offset(b + offset)
                self.buffer.apply_tag_by_name("bold", start, stop)
        for items in markup_pos["I"]:
            if len(items) == 3:
                (a, attributes, b) = items
                start = self.buffer.get_iter_at_offset(a + offset)
                stop = self.buffer.get_iter_at_offset(b + offset)
                self.buffer.apply_tag_by_name("italic", start, stop)
        for items in markup_pos["U"]:
            if len(items) == 3:
                (a, attributes, b) = items
                start = self.buffer.get_iter_at_offset(a + offset)
                stop = self.buffer.get_iter_at_offset(b + offset)
                self.buffer.apply_tag_by_name("underline", start, stop)
        for items in markup_pos["A"]:
            if len(items) == 3:
                (a, attributes, b) = items
                start = self.buffer.get_iter_at_offset(a + offset)
                stop = self.buffer.get_iter_at_offset(b + offset)
                if "href" in attributes:
                    url = attributes["href"]
                    self.link_region(start, stop, "URL", url)  # tooltip?
                elif "wiki" in attributes:
                    url = attributes["wiki"]
                    self.link_region(start, stop, "WIKI", url)  # tooltip?
                else:
                    LOG.warning("warning: no url on link: '%s'", text[start, stop])

    def link_region(self, start, stop, link_type, url):
        link_data = (LinkTag(self.buffer), link_type, url, url)
        self._tags.append(link_data)
        self.buffer.apply_tag(link_data[0], start, stop)

    def set_use_markup(self, value):
        if self.use_markup == value:
            return
        self.use_markup = value
        if value:
            self.buffer.create_tag("bold", weight=Pango.Weight.HEAVY)
            self.buffer.create_tag("italic", style=Pango.Style.ITALIC)
            self.buffer.create_tag("underline", underline=Pango.Underline.SINGLE)
            self.buffer.create_tag("fixed", font="monospace")
        else:
            tag_table = self.buffer.get_tag_table()
            tag_table.foreach(lambda tag, data: tag_table.remove(tag))

    def set_text(self, text, scroll_to="start"):
        self.buffer.set_text("")
        self.text_length = 0
        self.append_text(text, scroll_to)
        self.buffer.reset()

    def get_container_widget(self):
        raise NotImplementedError

    def add_gui_option(self, option):
        """
        Add an option to the GUI gramplet.
        """
        return make_gui_option(option, self.dbstate, self.uistate, self.track)

    def make_gui_options(self):
        if not self.pui:
            return
        # BEGIN WORKAROUND:
        # This is necessary because gtk doesn't redisplay these widgets
        # correctly so we replace them with new ones
        self.pui.save_options()
        self.pui.update_options = {}
        self.pui.option_order = []
        self.pui.build_options()
        # END WORKAROUND
        if len(self.pui.option_order) == 0:
            return
        frame = Gtk.Frame()
        topbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        hbox = Gtk.Grid()
        hbox.set_column_spacing(5)
        topbox.pack_start(hbox, False, False, 0)
        row = 0
        for item in self.pui.option_order:
            label = Gtk.Label(label=item + COLON)
            label.set_halign(Gtk.Align.END)
            hbox.attach(label, 0, row, 1, 1)
            # put Widget next to label
            hbox.attach(self.pui.option_dict[item][0], 1, row, 1, 1)
            row += 1
        save_button = Gtk.Button.new_with_mnemonic(_("_Save"))
        topbox.pack_end(save_button, False, False, 0)
        save_button.connect("clicked", self.pui.save_update_options)
        frame.add(topbox)
        frame.show_all()
        return frame

    def link(self, text, link_type, data, size=None, tooltip=None):
        buffer = self.buffer
        iter = buffer.get_end_iter()
        offset = buffer.get_char_count()
        self.append_text(text)
        start = buffer.get_iter_at_offset(offset)
        end = buffer.get_end_iter()
        link_data = (LinkTag(buffer), link_type, data, tooltip)
        if size:
            link_data[0].set_property("size-points", size)
        self._tags.append(link_data)
        buffer.apply_tag(link_data[0], start, end)

    def on_motion(self, view, event):
        buffer_location = view.window_to_buffer_coords(
            Gtk.TextWindowType.TEXT, int(event.x), int(event.y)
        )
        iter = view.get_iter_at_location(*buffer_location)
        if isinstance(iter, tuple):
            iter = iter[1]
        cursor = self.standard_cursor
        ttip = None
        for tag, link_type, handle, tooltip in self._tags:
            if iter.has_tag(tag):
                tag.set_property("underline", Pango.Underline.SINGLE)
                cursor = self.link_cursor
                ttip = tooltip
            else:
                tag.set_property("underline", Pango.Underline.NONE)
        view.get_window(Gtk.TextWindowType.TEXT).set_cursor(cursor)
        if ttip:
            self.scrolledwindow.set_tooltip_text(ttip)
        elif self.tooltips_text:
            self.scrolledwindow.set_tooltip_text(self.tooltips_text)
        return False  # handle event further, if necessary

    def on_button_press(self, view, event):
        # pylint: disable-msg=W0212
        buffer_location = view.window_to_buffer_coords(
            Gtk.TextWindowType.TEXT, int(event.x), int(event.y)
        )
        iter = view.get_iter_at_location(*buffer_location)
        if isinstance(iter, tuple):
            iter = iter[1]
        for tag, link_type, handle, tooltip in self._tags:
            if iter.has_tag(tag):
                if link_type == "Person":
                    if not self.dbstate.db.has_person_handle(handle):
                        return True
                    person = self.dbstate.db.get_person_from_handle(handle)
                    if person is not None:
                        if event.button == 1:  # left mouse
                            if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                                try:
                                    EditPerson(self.dbstate, self.uistate, [], person)
                                    return True  # handled event
                                except WindowActiveError:
                                    pass
                            elif event.type == Gdk.EventType.BUTTON_PRESS:
                                self.uistate.set_active(handle, "Person")
                                return True  # handled event
                        elif is_right_click(event):
                            # FIXME: add a popup menu with options
                            try:
                                EditPerson(self.dbstate, self.uistate, [], person)
                                return True  # handled event
                            except WindowActiveError:
                                pass
                elif link_type == "Surname":
                    if event.button == 1:  # left mouse
                        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                            run_quick_report_by_name(
                                self.dbstate, self.uistate, "samesurnames", handle
                            )
                    return True
                elif link_type == "Given":
                    if event.button == 1:  # left mouse
                        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                            run_quick_report_by_name(
                                self.dbstate, self.uistate, "samegivens_misc", handle
                            )
                    return True
                elif link_type == "Filter":
                    if event.button == 1:  # left mouse
                        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                            run_quick_report_by_name(
                                self.dbstate, self.uistate, "filterbyname", handle
                            )
                    return True
                elif link_type == "URL":
                    if event.button == 1:  # left mouse
                        display_url(handle)
                    return True
                elif link_type == "WIKI":
                    if event.button == 1:  # left mouse
                        handle = handle.replace(" ", "_")
                        if "#" in handle:
                            page, section = handle.split("#", 1)
                            display_help(page, section)
                        else:
                            display_help(handle)
                    return True
                elif link_type == "Family":
                    if not self.dbstate.db.has_family_handle(handle):
                        return True
                    family = self.dbstate.db.get_family_from_handle(handle)
                    if family is not None:
                        if event.button == 1:  # left mouse
                            if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                                try:
                                    EditFamily(self.dbstate, self.uistate, [], family)
                                    return True  # handled event
                                except WindowActiveError:
                                    pass
                            elif event.type == Gdk.EventType.BUTTON_PRESS:
                                self.uistate.set_active(handle, "Family")
                                return True  # handle event
                        elif is_right_click(event):
                            # FIXME: add a popup menu with options
                            try:
                                EditFamily(self.dbstate, self.uistate, [], family)
                                return True  # handled event
                            except WindowActiveError:
                                pass
                elif link_type == "PersonList":
                    if event.button == 1:  # left mouse
                        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                            run_quick_report_by_name(
                                self.dbstate,
                                self.uistate,
                                "filterbyname",
                                "list of people",
                                handles=handle,
                            )
                    return True
                elif link_type == "Attribute":
                    if event.button == 1:  # left mouse
                        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                            run_quick_report_by_name(
                                self.dbstate, self.uistate, "attribute_match", handle
                            )
                    return True
                else:  # overzealous l10n while setting the link?
                    logging.warning(
                        "Unknown link type %s, %s" % (link_type, RuntimeWarning)
                    )
        return False  # did not handle event

    def set_has_data(self, value):
        if isinstance(self.pane, Gtk.Notebook):
            if self.pane.page_num(self) != -1:
                label = self.pane.get_tab_label(self)
                label.set_has_data(value)


class GridGramplet(GuiGramplet):
    """
    Class that handles the plugin interfaces for the GrampletView.
    """

    TARGET_TYPE_FRAME = 80
    LOCAL_DRAG_TYPE = "GRAMPLET"
    LOCAL_DRAG_TARGET = Gtk.TargetEntry.new(LOCAL_DRAG_TYPE, 0, TARGET_TYPE_FRAME)

    def __init__(self, pane, dbstate, uistate, title, **kwargs):
        """
        Internal constructor for GUI portion of a gramplet.
        """
        GuiGramplet.__init__(self, pane, dbstate, uistate, title, **kwargs)

        self.xml = Glade()
        self.gvwin = self.xml.toplevel
        self.mainframe = self.xml.get_object("gvgramplet")
        self.gvwin.remove(self.mainframe)

        self.textview = self.xml.get_object("gvtextview")
        self.buffer = UndoableBuffer()
        self.text_length = 0
        self.textview.set_buffer(self.buffer)
        self.textview.connect("key-press-event", self.on_key_press_event)
        # self.buffer = self.textview.get_buffer()
        self.scrolledwindow = self.xml.get_object("gvscrolledwindow")
        self.scrolledwindow.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )
        self.vboxtop = self.xml.get_object("vboxtop")
        self.titlelabel = self.xml.get_object("gvtitle")
        self.titlelabel.get_children()[0].set_text("<b><i>%s</i></b>" % self.title)
        self.titlelabel.get_children()[0].set_use_markup(True)
        self.titlelabel.connect("clicked", self.edit_title)
        self.titlelabel_entry = None
        self.gvclose = self.xml.get_object("gvclose")
        self.gvclose.connect("clicked", self.close)
        self.gvstate = self.xml.get_object("gvstate")
        self.gvstate.connect("clicked", self.change_state)
        self.gvproperties = self.xml.get_object("gvproperties")
        self.gvproperties.connect("clicked", self.set_properties)
        self.xml.get_object("gvcloseimage").set_from_icon_name(
            "window-close", Gtk.IconSize.MENU
        )
        self.xml.get_object("gvstateimage").set_from_icon_name(
            "list-remove", Gtk.IconSize.MENU
        )
        self.xml.get_object("gvpropertiesimage").set_from_icon_name(
            "document-properties", Gtk.IconSize.MENU
        )

        # source:
        drag = self.gvproperties
        drag.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [GridGramplet.LOCAL_DRAG_TARGET],
            Gdk.DragAction.COPY,
        )

        # default tooltip
        msg = _("Drag Properties Button to move and click it for setup")
        if not self.tooltips_text:
            self.set_tooltip(msg)

    def edit_title(self, widget):
        """
        Edit the title in the GUI.
        """
        parent = widget.get_parent()
        widget.hide()
        if self.titlelabel_entry is None:
            self.titlelabel_entry = Gtk.Entry()
            parent = widget.get_parent()
            parent.pack_end(self.titlelabel_entry, True, True, 0)
            self.titlelabel_entry.connect("focus-out-event", self.edit_title_done)
            self.titlelabel_entry.connect("activate", self.edit_title_done)
            self.titlelabel_entry.connect("key-press-event", self.edit_title_keypress)
        self.titlelabel_entry.set_text(widget.get_children()[0].get_text())
        self.titlelabel_entry.show()
        self.titlelabel_entry.grab_focus()
        return True

    def edit_title_keypress(self, widget, event):
        """
        Edit the title, handle escape.
        """
        if event.type == Gdk.EventType.KEY_PRESS:
            if event.keyval == Gdk.KEY_Escape:
                self.titlelabel.show()
                widget.hide()

    def edit_title_done(self, widget, event=None):
        """
        Edit title in GUI, finishing callback.
        """
        result = self.set_title(widget.get_text())
        if result:  # if ok to set title to that
            self.titlelabel.show()
            widget.hide()
        return False  # Return False for gtk requirement

    def close(self, *obj):
        """
        Remove (delete) the gramplet from view.
        """
        if self.gstate == "detached":
            return
        self.gstate = "closed"
        self.pane.closed_gramplets.append(self)
        self.mainframe.get_parent().remove(self.mainframe)

    def detach(self):
        """
        Detach the gramplet from the GrampletView, and open in own window.
        """
        # hide buttons:
        # self.set_state("detached")
        self.pane.detached_gramplets.append(self)
        # make a window, and attach it there
        self.detached_window = GrampletWindow(self)

    def set_state(self, state):
        """
        Set the state of a gramplet.
        """
        oldstate = self.gstate
        self.gstate = state
        if state == "minimized":
            self.scrolledwindow.hide()
            self.xml.get_object("gvstateimage").set_from_icon_name(
                "list-add", Gtk.IconSize.MENU
            )
            column = self.mainframe.get_parent()  # column
            expand, fill, padding, pack = column.query_child_packing(self.mainframe)
            column.set_child_packing(self.mainframe, False, fill, padding, pack)
        else:
            self.scrolledwindow.show()
            self.xml.get_object("gvstateimage").set_from_icon_name(
                "list-remove", Gtk.IconSize.MENU
            )
            column = self.mainframe.get_parent()  # column
            expand, fill, padding, pack = column.query_child_packing(self.mainframe)
            column.set_child_packing(self.mainframe, self.expand, fill, padding, pack)
            if self.pui and self.pui.dirty:
                self.pui.update()

    def change_state(self, obj):
        """
        Change the state of a gramplet.
        """
        if self.gstate == "detached":
            pass  # don't change if detached
        else:
            if self.gstate == "maximized":
                self.set_state("minimized")
            else:
                self.set_state("maximized")

    def set_properties(self, obj):
        """
        Set the properties of a gramplet.
        """
        if self.gstate == "detached":
            pass
        else:
            self.detach()
        return
        self.expand = not self.expand
        if self.gstate == "maximized":
            column = self.mainframe.get_parent()  # column
            expand, fill, padding, pack = column.query_child_packing(self.mainframe)
            column.set_child_packing(self.mainframe, self.expand, fill, padding, pack)

    def get_source_widget(self):
        """
        Hack to allow us to send this object to the drop_widget
        method as a context.
        """
        return self.gvproperties

    def get_container_widget(self):
        return self.scrolledwindow

    def get_title(self):
        return self.title

    def set_height(self, height):
        self.height = height
        self.scrolledwindow.set_size_request(-1, self.height)
        self.set_state(self.gstate)

    def get_height(self):
        return self.height

    def get_detached_height(self):
        return self.detached_height

    def get_detached_width(self):
        return self.detached_width

    def set_detached_height(self, height):
        self.detached_height = height

    def set_detached_width(self, width):
        self.detached_width = width

    def get_expand(self):
        return self.expand

    def set_expand(self, value):
        self.expand = value
        self.scrolledwindow.set_size_request(-1, self.height)
        self.set_state(self.gstate)

    def set_title(self, new_title, set_override=True):
        # can't do it if already titled that way
        if self.title == new_title:
            return True
        if new_title in self.pane.gramplet_map or new_title != escape(
            new_title
        ):  # avoid XML specific characters
            return False
        if set_override:
            self.title_override = True
        del self.pane.gramplet_map[self.title]
        self.title = new_title
        if self.detached_window:
            self.detached_window.window.set_title(
                "%s %s - Gramps" % (new_title, _("Gramplet"))
            )
        self.pane.gramplet_map[self.title] = self
        self.titlelabel.get_children()[0].set_text("<b><i>%s</i></b>" % self.title)
        self.titlelabel.get_children()[0].set_use_markup(True)
        return True


class GrampletPane(Gtk.ScrolledWindow):
    def __init__(self, configfile, pageview, dbstate, uistate, **kwargs):
        self._config = Configuration(self)
        self.track = []
        Gtk.ScrolledWindow.__init__(self)
        self.configfile = os.path.join(VERSION_DIR, "%s.ini" % configfile)
        # default for new user; may be overridden in config:
        self.column_count = kwargs.get("column_count", 2)
        # width of window, if sidebar; may be overridden in config:
        self.pane_position = kwargs.get("pane_position", -1)
        self.pane_orientation = kwargs.get("pane_orientation", "horizontal")
        self.splitview = kwargs.get("splitview", None)
        self.default_gramplets = kwargs.get(
            "default_gramplets", ["Top Surnames", "Welcome"]
        )
        self.dbstate = dbstate
        self.uistate = uistate
        self.pageview = pageview
        self.pane = self
        self._popup_xy = None
        self.at_popup_action = None
        self.at_popup_menu = None
        user_gramplets = self.load_gramplets()
        # build the GUI:
        msg = _("Right click to add gramplets")
        self.set_tooltip_text(msg)
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.eventb = Gtk.EventBox()
        self.hbox = Gtk.Box(homogeneous=True)
        self.eventb.add(self.hbox)
        self.add(self.eventb)
        self.set_kinetic_scrolling(True)
        self.set_capture_button_press(True)
        # Set up drag and drop
        self.drag_dest_set(
            Gtk.DestDefaults.MOTION
            | Gtk.DestDefaults.HIGHLIGHT
            | Gtk.DestDefaults.DROP,
            [GridGramplet.LOCAL_DRAG_TARGET],
            Gdk.DragAction.COPY,
        )
        self.connect("drag_drop", self.drop_widget)
        self.eventb.connect("button-press-event", self._button_press)

        # Create the columns:
        self.columns = []
        for i in range(self.column_count):
            self.columns.append(Gtk.Box(orientation=Gtk.Orientation.VERTICAL))
            self.hbox.pack_start(self.columns[-1], True, True, 0)
        # Load the gramplets
        self.gramplet_map = {}  # title->gramplet
        self.frame_map = {}  # frame->gramplet
        self.detached_gramplets = []  # list of detached gramplets
        self.closed_gramplets = []  # list of closed gramplets
        self.closed_opts = []  # list of closed options from ini file
        # get the user's gramplets from ~/.gramps/gramplets.ini
        # Load the user's gramplets:
        for name_opts in user_gramplets:
            if name_opts is None:
                continue
            (name, opts) = name_opts
            all_opts = get_gramplet_opts(name, opts)
            if "state" not in all_opts:
                all_opts["state"] = "maximized"
            if all_opts["state"] == "closed":
                self.gramplet_map[all_opts["title"]] = None  # save closed name
                self.closed_opts.append(all_opts)
                continue
            if "title" not in all_opts:
                all_opts["title"] = _("Untitled Gramplet")
                set_override = False
            else:
                set_override = True
            # May have to change title
            g = make_requested_gramplet(
                GridGramplet, self, all_opts, self.dbstate, self.uistate
            )
            if g:
                g.title_override = (
                    set_override  # to continue to override, when this is saved
                )
                # make a unique title:
                unique = g.get_title()
                cnt = 1
                while unique in self.gramplet_map:
                    unique = g.get_title() + ("-%d" % cnt)
                    cnt += 1
                g.set_title(unique, set_override=False)
                self.gramplet_map[unique] = g
                self.frame_map[str(g.mainframe)] = g
            else:
                LOG.warning("Can't make gramplet of type '%s'.", name)
        self.place_gramplets()

    def show_all(self):
        """
        This seems to be necessary to hide the hidden
        parts of a collapsed gramplet on main view.
        """
        super(GrampletPane, self).show_all()
        for gramplet in list(self.gramplet_map.values()):
            if gramplet.gstate == "minimized":
                gramplet.set_state("minimized")

    def set_state_all(self):
        """
        This seems to be necessary to hide the hidden
        parts of a collapsed gramplet on sidebars.
        """
        for gramplet in list(self.gramplet_map.values()):
            if gramplet.gstate in ["minimized", "maximized"]:
                gramplet.set_state(gramplet.gstate)

    def get_column_frame(self, column_num):
        if column_num < len(self.columns):
            return self.columns[column_num]
        else:
            return self.columns[-1]  # it was too big, so select largest

    def clear_gramplets(self):
        """
        Detach all of the mainframe gramplets from the columns.
        """
        gramplets = (g for g in self.gramplet_map.values() if g is not None)
        for gramplet in gramplets:
            if gramplet.gstate == "detached" or gramplet.gstate == "closed":
                continue
            column = gramplet.mainframe.get_parent()
            if column:
                column.remove(gramplet.mainframe)

    def place_gramplets(self, recolumn=False):
        """
        Place the gramplet mainframes in the columns.
        """
        gramplets = [g for g in self.gramplet_map.values() if g is not None]
        # put the gramplets where they go:
        # sort by row
        gramplets.sort(key=lambda x: x.row)
        rows = [0] * max(self.column_count, 1)
        for cnt, gramplet in enumerate(gramplets):
            # see if the user wants this in a particular location:
            # and if there are that many columns
            if gramplet.column >= 0 and gramplet.column < self.column_count:
                pos = gramplet.column
            else:
                # else, spread them out:
                pos = cnt % self.column_count
            gramplet.column = pos
            gramplet.row = rows[gramplet.column]
            rows[gramplet.column] += 1
            if recolumn and (
                gramplet.gstate == "detached" or gramplet.gstate == "closed"
            ):
                continue
            if gramplet.gstate == "minimized":
                self.columns[pos].pack_start(gramplet.mainframe, False, True, 0)
            else:
                self.columns[pos].pack_start(
                    gramplet.mainframe, gramplet.expand, True, 0
                )
            # set height on gramplet.scrolledwindow here:
            gramplet.scrolledwindow.set_size_request(-1, gramplet.height)
            # Can't minimize here, because Gramps calls show_all later:
            # if gramplet.gstate == "minimized": # starts max, change to min it
            #    gramplet.set_state("minimized") # minimize it
            # set minimized is called in page subclass hack (above)
            if gramplet.gstate == "detached":
                gramplet.detach()
            elif gramplet.gstate == "closed":
                gramplet.close()

    def load_gramplets(self):
        retval = []
        filename = self.configfile
        if filename and os.path.exists(filename):
            cp = configparser.ConfigParser(strict=False)
            try:
                cp.read(filename, encoding="utf-8")
            except Exception as err:
                LOG.warning(
                    "Failed to load gramplets from %s because %s", filename, str(err)
                )
                return [None]
            for sec in cp.sections():
                if sec == "Gramplet View Options":
                    if "column_count" in cp.options(sec):
                        self.column_count = int(cp.get(sec, "column_count"))
                    if "pane_position" in cp.options(sec):
                        self.pane_position = int(cp.get(sec, "pane_position"))
                    if "pane_orientation" in cp.options(sec):
                        self.pane_orientation = cp.get(sec, "pane_orientation")
                else:
                    data = {}
                    for opt in cp.options(sec):
                        if opt.startswith("data["):
                            temp = data.get("data", {})
                            # temp.append(cp.get(sec, opt).strip())
                            pos = int(opt[5:-1])
                            temp[pos] = cp.get(sec, opt).strip()
                            data["data"] = temp
                        else:
                            data[opt] = cp.get(sec, opt).strip()
                    if "data" in data:
                        data["data"] = [
                            data["data"][key] for key in sorted(data["data"].keys())
                        ]
                    if "name" not in data:
                        data["name"] = "Unnamed Gramplet"
                        data["tname"] = _("Unnamed Gramplet")
                    retval.append((data["name"], data))  # name, opts
            clean_up(cp)
        else:
            # give defaults as currently known
            for name in self.default_gramplets:
                if name in AVAILABLE_GRAMPLETS():
                    retval.append((name, GET_AVAILABLE_GRAMPLETS(name)))
        return retval

    def save(self):
        if len(self.frame_map) + len(self.detached_gramplets) == 0:
            return  # something is the matter
        filename = self.configfile
        try:
            with open(filename, "w", encoding="utf-8") as fp:
                fp.write(";; Gramps gramplets file\n")
                fp.write(
                    ";; Automatically created at %s"
                    % time.strftime("%Y/%m/%d %H:%M:%S\n\n")
                )
                fp.write("[Gramplet View Options]\n")
                fp.write("column_count=%d\n" % self.column_count)
                fp.write("pane_position=%d\n" % self.pane_position)
                fp.write("pane_orientation=%s\n\n" % self.pane_orientation)
                # showing gramplets:
                for col in range(self.column_count):
                    row = 0
                    for gframe in self.columns[col]:
                        gramplet = self.frame_map[str(gframe)]
                        opts = get_gramplet_options_by_name(gramplet.gname)
                        if opts is not None:
                            base_opts = opts.copy()
                            for key in base_opts:
                                if key in gramplet.__dict__:
                                    base_opts[key] = gramplet.__dict__[key]
                            base_opts["state"] = gramplet.gstate
                            fp.write("[%s]\n" % gramplet.title)  # section
                            for key in base_opts:
                                if key == "content":
                                    continue
                                elif key == "tname":
                                    continue
                                elif key == "column":
                                    continue
                                elif key == "row":
                                    continue
                                elif key == "version":
                                    continue  # code, don't save
                                elif key == "gramps":
                                    continue  # code, don't save
                                elif key == "data":
                                    if not isinstance(base_opts["data"], (list, tuple)):
                                        fp.write("data[0]=%s\n" % base_opts["data"])
                                    else:
                                        cnt = 0
                                        for item in base_opts["data"]:
                                            # If we have a "%" in a string,
                                            # escape it by writing "%%"
                                            # to avoid InterpolationSyntaxError
                                            # in python configparser module.
                                            if isinstance(item, str):
                                                item = item.replace("%", "%%")
                                            fp.write("data[%d]=%s\n" % (cnt, item))
                                            cnt += 1
                                else:
                                    fp.write("%s=%s\n" % (key, base_opts[key]))
                            fp.write("column=%d\n" % col)
                            fp.write("row=%d\n\n" % row)
                        row += 1
                for gramplet in self.detached_gramplets:
                    opts = get_gramplet_options_by_name(gramplet.gname)
                    if opts is not None:
                        base_opts = opts.copy()
                        for key in base_opts:
                            if key in gramplet.__dict__:
                                base_opts[key] = gramplet.__dict__[key]
                        base_opts["state"] = gramplet.gstate
                        fp.write("[%s]\n" % gramplet.title)
                        for key in base_opts:
                            if key == "content":
                                continue
                            elif key == "title":
                                if "title_override" in base_opts:
                                    base_opts["title"] = base_opts["title_override"]
                                fp.write("title=%s\n" % base_opts[key])
                            elif key == "tname":
                                continue
                            elif key == "version":
                                continue  # code, don't save
                            elif key == "gramps":
                                continue  # code, don't save
                            elif key == "data":
                                if not isinstance(base_opts["data"], (list, tuple)):
                                    fp.write("data[0]=%s\n" % base_opts["data"])
                                else:
                                    cnt = 0
                                    for item in base_opts["data"]:
                                        fp.write("data[%d]=%s\n" % (cnt, item))
                                        cnt += 1
                            else:
                                fp.write("%s=%s\n" % (key, base_opts[key]))

        except IOError as err:
            LOG.warning(
                "Failed to open %s because $s; gramplets not saved", filename, str(err)
            )
            return

    def drop_widget(self, source, context, x, y, timedata):
        """
        This is the destination method for handling drag and drop
        of a gramplet onto the main scrolled window.
        Also used for adding new gramplets, then context should be GridGramplet
        """
        button = None
        if isinstance(context, Gdk.DragContext):
            button = Gtk.drag_get_source_widget(context)
        else:
            button = context.get_source_widget()
        if button:
            hbox = button.get_parent()
            mframe = hbox.get_parent()
            mainframe = mframe.get_parent()  # actually a vbox
        rect = source.get_allocation()
        sx, sy = rect.width, rect.height
        # Convert to LTR co-ordinates when using RTL locale
        if source.get_direction() == Gtk.TextDirection.RTL:
            x = sx - x
        # first, find column:
        col = 0
        for i, column in enumerate(self.columns):
            if x < (sx / len(self.columns) * (i + 1)):
                col = i
                break
        if button:
            fromcol = mainframe.get_parent()
            if fromcol:
                fromcol.remove(mainframe)
        # now find where to insert in column:
        stack = []
        current_row = 0
        for gframe in self.columns[col]:
            gramplet = self.frame_map[str(gframe)]
            gramplet.row = current_row
            current_row += 1
            rect = gframe.get_allocation()
            if y < (rect.y + 15):  # starts at 0, this allows insert before
                self.columns[col].remove(gframe)
                stack.append(gframe)
        maingramplet = self.frame_map.get(str(mainframe), None)
        maingramplet.column = col
        maingramplet.row = current_row
        current_row += 1
        expand = maingramplet.gstate == "maximized" and maingramplet.expand
        self.columns[col].pack_start(mainframe, expand, True, 0)
        for gframe in stack:
            gramplet = self.frame_map[str(gframe)]
            gramplet.row = current_row
            current_row += 1
            expand = gramplet.gstate == "maximized" and gramplet.expand
            self.columns[col].pack_start(gframe, expand, True, 0)
        return True

    def set_columns(self, num):
        if num < 1:
            num = 1
        # clear the gramplets:
        self.clear_gramplets()
        # clear the columns:
        for column in self.columns:
            frame = column.get_parent()
            frame.remove(column)
            del column
        # create the new ones:
        self.column_count = num
        self.columns = []
        for i in range(self.column_count):
            self.columns.append(Gtk.Box(orientation=Gtk.Orientation.VERTICAL))
            self.columns[-1].show()
            self.hbox.pack_start(self.columns[-1], True, True, 0)
        # place the gramplets back in the new columns
        self.place_gramplets(recolumn=True)
        self.show()

    def restore_gramplet(self, name):
        ############### First kind: from current session
        for gramplet in self.closed_gramplets:
            if gramplet.title == name:
                # gramplet.gstate = "maximized"
                self.closed_gramplets.remove(gramplet)
                if self._popup_xy is not None:
                    self.drop_widget(
                        self, gramplet, self._popup_xy[0], self._popup_xy[1], 0
                    )
                else:
                    self.drop_widget(self, gramplet, 0, 0, 0)
                gramplet.set_state("maximized")
                return
        ################ Second kind: from options
        for opts in self.closed_opts:
            if opts["title"] == name:
                self.closed_opts.remove(opts)
                g = make_requested_gramplet(
                    GridGramplet, self, opts, self.dbstate, self.uistate
                )
                if g:
                    self.gramplet_map[opts["title"]] = g
                    self.frame_map[str(g.mainframe)] = g
                else:
                    LOG.warning("Can't make gramplet of type '%s'.", name)
        if g:
            gramplet = g
            gramplet.gstate = "maximized"
            if gramplet.column >= 0 and gramplet.column < len(self.columns):
                pos = gramplet.column
            else:
                pos = 0
            self.columns[pos].pack_start(gramplet.mainframe, expand=gramplet.expand)
            # set height on gramplet.scrolledwindow here:
            gramplet.scrolledwindow.set_size_request(-1, gramplet.height)
            ## now drop it in right place
            if self._popup_xy is not None:
                self.drop_widget(
                    self, gramplet, self._popup_xy[0], self._popup_xy[1], 0
                )
            else:
                self.drop_widget(self, gramplet, 0, 0, 0)

    def add_gramplet(self, tname):
        all_opts = get_gramplet_options_by_tname(tname)
        name = all_opts["name"]
        if all_opts is None:
            LOG.warning("Unknown gramplet type: '%s'; bad " "gramplets.ini file?", name)
            return
        if "title" not in all_opts:
            all_opts["title"] = "Untitled Gramplet"
        # uniqify titles:
        unique = all_opts["title"]
        cnt = 1
        while unique in self.gramplet_map:
            unique = all_opts["title"] + ("-%d" % cnt)
            cnt += 1
        all_opts["title"] = unique
        if all_opts["title"] not in self.gramplet_map:
            g = make_requested_gramplet(
                GridGramplet, self, all_opts, self.dbstate, self.uistate
            )
        if g:
            self.gramplet_map[all_opts["title"]] = g
            self.frame_map[str(g.mainframe)] = g
            gramplet = g
            if gramplet.column >= 0 and gramplet.column < len(self.columns):
                pos = gramplet.column
            else:
                pos = 0
            self.columns[pos].pack_start(gramplet.mainframe, gramplet.expand, True, 0)
            # set height on gramplet.scrolledwindow here:
            gramplet.scrolledwindow.set_size_request(-1, gramplet.height)
            ## now drop it in right place
            if self._popup_xy is not None:
                self.drop_widget(
                    self, gramplet, self._popup_xy[0], self._popup_xy[1], 0
                )
            else:
                self.drop_widget(self, gramplet, 0, 0, 0)
            if gramplet.pui:
                gramplet.pui.active = True
                gramplet.pui.update()
        else:
            LOG.warning("Can't make gramplet of type '%s'.", name)

    def _button_press(self, obj, event):
        ui_def = (
            """    <menu id="Popup">
        <submenu>
          <attribute name="action">win.AddGramplet</attribute>
          <attribute name="label" translatable="yes">Add a gramplet</attribute>
          %s
        </submenu>
        <submenu>
          <attribute name="action">win.RestoreGramplet</attribute>
          <attribute name="label" translatable="yes">"""
            """Restore a gramplet</attribute>
          %s
        </submenu>
        </menu>
        """
        )
        menuitem = (
            "<item>\n"
            '<attribute name="action">win.%s</attribute>\n'
            '<attribute name="label">%s</attribute>\n'
            "</item>\n"
        )

        if is_right_click(event):
            self._popup_xy = (event.x, event.y)
            uiman = self.uistate.uimanager
            actions = []
            r_menuitems = ""
            a_menuitems = ""
            plugs = [
                gplug
                for gplug in PLUGMAN.get_reg_gramplets()
                if gplug.navtypes == [] or "Dashboard" in gplug.navtypes
            ]
            plugs.sort(key=lambda x: x.name)
            for plug in plugs:
                action_name = valid_action_name(plug.id)
                a_menuitems += menuitem % (action_name, escape(plug.name))
                actions.append(
                    (action_name, make_callback(self.add_gramplet, plug.name))
                )
            names = [gramplet.title for gramplet in self.closed_gramplets]
            names.extend(opts["title"] for opts in self.closed_opts)
            names.sort()
            if len(names) > 0:
                for name in names:
                    # 'name' could be non-ASCII when in non-English language
                    # action names must be in ASCII, so use 'id' instead.
                    action_name = valid_action_name(str(id(name)))
                    r_menuitems += menuitem % (action_name, escape(name))
                    actions.append(
                        (action_name, make_callback(self.restore_gramplet, name))
                    )

            if self.at_popup_action:
                uiman.remove_ui(self.at_popup_menu)
                uiman.remove_action_group(self.at_popup_action)
            self.at_popup_action = ActionGroup("AtPopupActions", actions)
            uiman.insert_action_group(self.at_popup_action)
            self.at_popup_menu = uiman.add_ui_from_string(
                [ui_def % (a_menuitems, r_menuitems)]
            )
            uiman.update_menu()

            menu = uiman.get_widget("Popup")
            popup_menu = Gtk.Menu.new_from_model(menu)
            popup_menu.attach_to_widget(obj, None)
            popup_menu.show_all()
            popup_menu.popup_at_pointer(event)
            return True
        return False

    def set_inactive(self):
        for title in self.gramplet_map:
            if self.gramplet_map[title].pui:
                if self.gramplet_map[title].gstate != "detached":
                    self.gramplet_map[title].pui.active = False

    def set_active(self):
        for title in self.gramplet_map:
            if self.gramplet_map[title].pui:
                self.gramplet_map[title].pui.active = True
                if self.gramplet_map[title].pui.dirty:
                    if self.gramplet_map[title].gstate == "maximized":
                        self.gramplet_map[title].pui.update()

    def on_delete(self):
        gramplets = (g for g in self.gramplet_map.values() if g is not None)
        for gramplet in gramplets:
            # this is the only place where the gui runs user code directly
            if gramplet.pui:
                gramplet.pui.on_save()
        self.save()

    def can_configure(self):
        """
        See :class:`.PageView`

        :return: bool
        """
        return True

    def _get_configure_page_funcs(self):
        """
        Return a list of functions that create gtk elements to use in the
        notebook pages of the Configure dialog

        :return: list of functions
        """

        def generate_pages():
            return [self.config_panel] + [
                self.build_panel(gramplet)
                for gramplet in sorted(
                    list(self.gramplet_map.values()), key=lambda g: g.title
                )
                if gramplet.gstate != "closed"
            ]

        return generate_pages

    def get_columns(self):
        return self.column_count

    def config_panel(self, configdialog):
        """
        Function that builds the widget in the configuration dialog
        """
        grid = Gtk.Grid()
        grid.set_border_width(12)
        grid.set_column_spacing(6)
        grid.set_row_spacing(6)

        self._config.register(
            "Gramplet View Options.column_count",
            int,
            self.get_columns,  # pane
            self.set_columns,
        )  # pane

        configdialog.add_pos_int_entry(
            grid,
            _("Number of Columns"),
            0,
            "Gramplet View Options.column_count",
            self._config.set,
            config=self._config,
        )
        return _("Gramplet Layout"), grid

    def build_panel(self, gramplet):
        self._config.register(
            "%s.title" % gramplet.title, str, gramplet.get_title, gramplet.set_title
        )
        self._config.register(
            "%s.height" % gramplet.title, int, gramplet.get_height, gramplet.set_height
        )
        self._config.register(
            "%s.detached_height" % gramplet.title,
            int,
            gramplet.get_detached_height,
            gramplet.set_detached_height,
        )
        self._config.register(
            "%s.detached_width" % gramplet.title,
            int,
            gramplet.get_detached_width,
            gramplet.set_detached_width,
        )
        self._config.register(
            "%s.expand" % gramplet.title, bool, gramplet.get_expand, gramplet.set_expand
        )

        def gramplet_panel(configdialog):
            configdialog.window.set_size_request(600, -1)
            grid = Gtk.Grid()
            grid.set_border_width(12)
            grid.set_column_spacing(6)
            grid.set_row_spacing(6)
            # Title:
            configdialog.add_entry(
                grid,
                _("Title"),
                0,
                "%s.title" % gramplet.title,
                self._config.set,
                config=self._config,
            )
            # Expand to max height
            configdialog.add_checkbox(
                grid,
                _("Use maximum height available"),
                1,
                "%s.expand" % gramplet.title,
                config=self._config,
            )
            # Height
            configdialog.add_pos_int_entry(
                grid,
                _("Height if not maximized"),
                2,
                "%s.height" % gramplet.title,
                self._config.set,
                config=self._config,
            )
            # Options:
            options = gramplet.make_gui_options()
            if options:
                grid.attach(options, 1, 5, 3, 1)
            return gramplet.title, grid

        return gramplet_panel


class Configuration:
    """
    A config wrapper to redirect set/get to GrampletPane.
    """

    def __init__(self, pane):
        self.pane = pane
        self.data = {}

    def get(self, key):
        vtype, getter, setter = self.data[key]
        return getter()

    def set(self, widget, key):
        """
        Hooked to signal, it is widget, key.
        Hooked to config, it is key, widget
        """
        if key not in self.data:
            widget, key = key, widget
        vtype, getter, setter = self.data[key]
        if type(widget) == vtype:
            setter(widget)
        else:
            try:
                value = vtype(widget.get_text())
            except:
                return
            setter(value)

    def register(self, key, vtype, getter, setter):
        """
        register a key with type, getter, and setter methods.
        """
        self.data[key] = (vtype, getter, setter)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".")
LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.lib import NoteType, Source
from gramps.gen.db import DbTxn
from .editprimary import EditPrimary

from .displaytabs import (
    NoteTab,
    GalleryTab,
    SrcAttrEmbedList,
    CitationBackRefList,
    RepoEmbedList,
)
from ..widgets import MonitoredEntry, PrivacyButton, MonitoredTagList
from ..dialog import ErrorDialog
from ..glade import Glade
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("New_Source_dialog", "manual")

# -------------------------------------------------------------------------
#
# EditSource class
#
# -------------------------------------------------------------------------


class EditSource(EditPrimary):
    def __init__(self, dbstate, uistate, track, source, callback=None):
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            source,
            dbstate.db.get_source_from_handle,
            dbstate.db.get_source_from_gramps_id,
            callback,
        )

    def empty_object(self):
        return Source()

    def get_menu_title(self):
        title = self.obj.get_title()
        if title:
            title = _("Source") + ": " + title
        else:
            title = _("New Source")
        return title

    def _local_init(self):
        assert self.obj

        self.glade = Glade()
        self.set_window(self.glade.toplevel, None, self.get_menu_title())
        self.setup_configs("interface.source", 600, 450)

    def _connect_signals(self):
        self.define_ok_button(self.glade.get_object("ok"), self.save)
        self.define_cancel_button(self.glade.get_object("cancel"))
        self.define_help_button(
            self.glade.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("source-rebuild", self._do_close)
        self._add_db_signal("source-delete", self.check_for_close)

    def _setup_fields(self):
        self.author = MonitoredEntry(
            self.glade.get_object("author"),
            self.obj.set_author,
            self.obj.get_author,
            self.db.readonly,
        )

        self.pubinfo = MonitoredEntry(
            self.glade.get_object("pubinfo"),
            self.obj.set_publication_info,
            self.obj.get_publication_info,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.glade.get_object("gid"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.glade.get_object("tag_label"),
            self.glade.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.priv = PrivacyButton(
            self.glade.get_object("private"), self.obj, self.db.readonly
        )

        self.abbrev = MonitoredEntry(
            self.glade.get_object("abbrev"),
            self.obj.set_abbreviation,
            self.obj.get_abbreviation,
            self.db.readonly,
        )

        self.title = MonitoredEntry(
            self.glade.get_object("source_title"),
            self.obj.set_title,
            self.obj.get_title,
            self.db.readonly,
        )

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "source_editor_notes",
            self.get_menu_title(),
            NoteType.SOURCE,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.gallery_tab = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_media_list(),
        )
        self._add_tab(notebook, self.gallery_tab)
        self.track_ref_for_deletion("gallery_tab")

        self.attr_tab = SrcAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_attribute_list(),
            "source_editor_attributes",
        )
        self._add_tab(notebook, self.attr_tab)
        self.track_ref_for_deletion("attr_tab")

        self.repo_tab = RepoEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_reporef_list(),
            "source_editor_repositories",
        )
        self._add_tab(notebook, self.repo_tab)
        self.track_ref_for_deletion("repo_tab")

        self.backref_list = CitationBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.obj.handle),
            "source_editor_references",
        )
        self.backref_tab = self._add_tab(notebook, self.backref_list)
        self.track_ref_for_deletion("backref_tab")
        self.track_ref_for_deletion("backref_list")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.glade.get_object("vbox").pack_start(notebook, True, True, 0)

    def build_menu_names(self, source):
        return (_("Edit Source"), self.get_menu_title())

    def save(self, *obj):
        self.ok_button.set_sensitive(False)
        if self.object_is_empty():
            ErrorDialog(
                _("Cannot save source"),
                _(
                    "No data exists for this source. Please "
                    "enter data or cancel the edit."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        (uses_dupe_id, id) = self._uses_duplicate_id()
        if uses_dupe_id:
            prim_object = self.get_from_gramps_id(id)
            name = prim_object.get_title()
            msg1 = _("Cannot save source. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used by '"
                "%(prim_object)s'. Please enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": id, "prim_object": name}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        if not self.obj.handle:
            with DbTxn(_("Add Source (%s)") % self.obj.get_title(), self.db) as trans:
                self.db.add_source(self.obj, trans)
        else:
            if self.data_has_changed():
                with DbTxn(
                    _("Edit Source (%s)") % self.obj.get_title(), self.db
                ) as trans:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_source_gramps_id())
                    self.db.commit_source(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj)
